# Copyright (C) 2019 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from hamcrest import ( assert_that,
                       empty )
from unittest import TestCase
from ycm import signature_help as sh


class SignatureHelpTest( TestCase ):
  def test_MakeSignatureHelpBuffer_Empty( self ):
    assert_that( sh._MakeSignatureHelpBuffer( {} ), empty() )
    assert_that( sh._MakeSignatureHelpBuffer( {
      'activeSignature': 0,
      'activeParameter': 0,
      'signatures': []
    } ), empty() )
    assert_that( sh._MakeSignatureHelpBuffer( {
      'activeSignature': 0,
      'activeParameter': 0,
    } ), empty() )
    assert_that( sh._MakeSignatureHelpBuffer( {
      'signatures': []
    } ), empty() )



# Copyright (C) 2013 Google Inc.
#               2020 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
from hamcrest import assert_that, equal_to
from unittest import TestCase
from unittest.mock import patch

from ycm.tests.test_utils import MockVimModule
vim_mock = MockVimModule()
from ycm import base


@contextlib.contextmanager
def MockCurrentFiletypes( filetypes = [ '' ] ):
  with patch( 'ycm.vimsupport.CurrentFiletypes', return_value = filetypes ):
    yield


@contextlib.contextmanager
def MockCurrentColumnAndLineContents( column, line_contents ):
  with patch( 'ycm.vimsupport.CurrentColumn', return_value = column ):
    with patch( 'ycm.vimsupport.CurrentLineContents',
                return_value = line_contents ):
      yield


@contextlib.contextmanager
def MockTextAfterCursor( text ):
  with patch( 'ycm.vimsupport.TextAfterCursor', return_value = text ):
    yield


class BaseTest( TestCase ):
  def test_AdjustCandidateInsertionText_Basic( self ):
    with MockTextAfterCursor( 'bar' ):
      assert_that( [ { 'word': 'foo',    'abbr': 'foobar' } ],
                   equal_to( base.AdjustCandidateInsertionText( [
                               { 'word': 'foobar', 'abbr': '' } ] ) ) )


  def test_AdjustCandidateInsertionText_ParenInTextAfterCursor( self ):
    with MockTextAfterCursor( 'bar(zoo' ):
      assert_that( [ { 'word': 'foo',    'abbr': 'foobar' } ],
                   equal_to( base.AdjustCandidateInsertionText( [
                               { 'word': 'foobar', 'abbr': '' } ] ) ) )


  def test_AdjustCandidateInsertionText_PlusInTextAfterCursor( self ):
    with MockTextAfterCursor( 'bar+zoo' ):
      assert_that( [ { 'word': 'foo',    'abbr': 'foobar' } ],
                   equal_to( base.AdjustCandidateInsertionText( [
                               { 'word': 'foobar', 'abbr': '' } ] ) ) )


  def test_AdjustCandidateInsertionText_WhitespaceInTextAfterCursor( self ):
    with MockTextAfterCursor( 'bar zoo' ):
      assert_that( [ { 'word': 'foo',    'abbr': 'foobar' } ],
                   equal_to( base.AdjustCandidateInsertionText( [
                               { 'word': 'foobar', 'abbr': '' } ] ) ) )


  def test_AdjustCandidateInsertionText_MoreThanWordMatchingAfterCursor( self ):
    with MockTextAfterCursor( 'bar.h' ):
      assert_that( [ { 'word': 'foo', 'abbr': 'foobar.h' } ],
                   equal_to( base.AdjustCandidateInsertionText( [
                               { 'word': 'foobar.h', 'abbr': '' } ] ) ) )

    with MockTextAfterCursor( 'bar(zoo' ):
      assert_that( [ { 'word': 'foo', 'abbr': 'foobar(zoo' } ],
                   equal_to( base.AdjustCandidateInsertionText( [
                               { 'word': 'foobar(zoo', 'abbr': '' } ] ) ) )


  def test_AdjustCandidateInsertionText_NotSuffix( self ):
    with MockTextAfterCursor( 'bar' ):
      assert_that( [ { 'word': 'foofoo', 'abbr': 'foofoo' } ],
                   equal_to( base.AdjustCandidateInsertionText( [
                     { 'word': 'foofoo', 'abbr': '' } ] ) ) )


  def test_AdjustCandidateInsertionText_NothingAfterCursor( self ):
    with MockTextAfterCursor( '' ):
      assert_that( [ { 'word': 'foofoo', 'abbr': '' },
                     { 'word': 'zobar',  'abbr': '' } ],
                   equal_to( base.AdjustCandidateInsertionText( [
                     { 'word': 'foofoo', 'abbr': '' },
                     { 'word': 'zobar',  'abbr': '' } ] ) ) )


  def test_AdjustCandidateInsertionText_MultipleStrings( self ):
    with MockTextAfterCursor( 'bar' ):
      assert_that( [ { 'word': 'foo',    'abbr': 'foobar' },
                     { 'word': 'zo',     'abbr': 'zobar' },
                     { 'word': 'q',      'abbr': 'qbar' },
                     { 'word': '',       'abbr': 'bar' }, ],
                   equal_to( base.AdjustCandidateInsertionText( [
                     { 'word': 'foobar', 'abbr': '' },
                     { 'word': 'zobar',  'abbr': '' },
                     { 'word': 'qbar',   'abbr': '' },
                     { 'word': 'bar',    'abbr': '' } ] ) ) )


  def test_AdjustCandidateInsertionText_DontTouchAbbr( self ):
    with MockTextAfterCursor( 'bar' ):
      assert_that( [ { 'word': 'foo',    'abbr': '1234' } ],
                   equal_to( base.AdjustCandidateInsertionText( [
                     { 'word': 'foobar', 'abbr': '1234' } ] ) ) )


  def test_AdjustCandidateInsertionText_NoAbbr( self ):
    with MockTextAfterCursor( 'bar' ):
      assert_that( [ { 'word': 'foo', 'abbr': 'foobar' } ],
                   equal_to( base.AdjustCandidateInsertionText( [
                     { 'word': 'foobar' } ] ) ) )


  def test_OverlapLength_Basic( self ):
    assert_that( 3, equal_to( base.OverlapLength( 'foo bar', 'bar zoo' ) ) )
    assert_that( 3, equal_to( base.OverlapLength( 'foobar', 'barzoo' ) ) )


  def test_OverlapLength_BasicWithUnicode( self ):
    assert_that( 3, equal_to( base.OverlapLength( 'bar fäö', 'fäö bar' ) ) )
    assert_that( 3, equal_to( base.OverlapLength( 'zoofäö', 'fäözoo' ) ) )


  def test_OverlapLength_OneCharOverlap( self ):
    assert_that( 1, equal_to( base.OverlapLength( 'foo b', 'b zoo' ) ) )


  def test_OverlapLength_SameStrings( self ):
    assert_that( 6, equal_to( base.OverlapLength( 'foobar', 'foobar' ) ) )


  def test_OverlapLength_Substring( self ):
    assert_that( 6, equal_to( base.OverlapLength( 'foobar', 'foobarzoo' ) ) )
    assert_that( 6, equal_to( base.OverlapLength( 'zoofoobar', 'foobar' ) ) )


  def test_OverlapLength_LongestOverlap( self ):
    assert_that( 7, equal_to( base.OverlapLength( 'bar foo foo',
                                                  'foo foo bar' ) ) )


  def test_OverlapLength_EmptyInput( self ):
    assert_that( 0, equal_to( base.OverlapLength( '', 'goobar' ) ) )
    assert_that( 0, equal_to( base.OverlapLength( 'foobar', '' ) ) )
    assert_that( 0, equal_to( base.OverlapLength( '', '' ) ) )


  def test_OverlapLength_NoOverlap( self ):
    assert_that( 0, equal_to( base.OverlapLength( 'foobar', 'goobar' ) ) )
    assert_that( 0, equal_to( base.OverlapLength( 'foobar', '(^($@#$#@' ) ) )
    assert_that( 0, equal_to( base.OverlapLength( 'foo bar zoo',
                                                  'foo zoo bar' ) ) )


  def test_LastEnteredCharIsIdentifierChar_Basic( self ):
    with MockCurrentFiletypes():
      with MockCurrentColumnAndLineContents( 3, 'abc' ):
        assert_that( base.LastEnteredCharIsIdentifierChar() )

      with MockCurrentColumnAndLineContents( 2, 'abc' ):
        assert_that( base.LastEnteredCharIsIdentifierChar() )

      with MockCurrentColumnAndLineContents( 1, 'abc' ):
        assert_that( base.LastEnteredCharIsIdentifierChar() )


  def test_LastEnteredCharIsIdentifierChar_FiletypeHtml( self ):
    with MockCurrentFiletypes( [ 'html' ] ):
      with MockCurrentColumnAndLineContents( 3, 'ab-' ):
        assert_that( base.LastEnteredCharIsIdentifierChar() )


  def test_LastEnteredCharIsIdentifierChar_ColumnIsZero( self ):
    with MockCurrentColumnAndLineContents( 0, 'abc' ):
      assert_that( not base.LastEnteredCharIsIdentifierChar() )


  def test_LastEnteredCharIsIdentifierChar_LineEmpty( self ):
    with MockCurrentFiletypes():
      with MockCurrentColumnAndLineContents( 3, '' ):
        assert_that( not base.LastEnteredCharIsIdentifierChar() )

      with MockCurrentColumnAndLineContents( 0, '' ):
        assert_that( not base.LastEnteredCharIsIdentifierChar() )


  def test_LastEnteredCharIsIdentifierChar_NotIdentChar( self ):
    with MockCurrentFiletypes():
      with MockCurrentColumnAndLineContents( 3, 'ab;' ):
        assert_that( not base.LastEnteredCharIsIdentifierChar() )

      with MockCurrentColumnAndLineContents( 1, ';' ):
        assert_that( not base.LastEnteredCharIsIdentifierChar() )

      with MockCurrentColumnAndLineContents( 3, 'ab-' ):
        assert_that( not base.LastEnteredCharIsIdentifierChar() )


  def test_LastEnteredCharIsIdentifierChar_Unicode( self ):
    with MockCurrentFiletypes():
      # CurrentColumn returns a byte offset and character ø is 2 bytes length.
      with MockCurrentColumnAndLineContents( 5, 'føo(' ):
        assert_that( not base.LastEnteredCharIsIdentifierChar() )

      with MockCurrentColumnAndLineContents( 4, 'føo(' ):
        assert_that( base.LastEnteredCharIsIdentifierChar() )

      with MockCurrentColumnAndLineContents( 3, 'føo(' ):
        assert_that( base.LastEnteredCharIsIdentifierChar() )

      with MockCurrentColumnAndLineContents( 1, 'føo(' ):
        assert_that( base.LastEnteredCharIsIdentifierChar() )


  def test_CurrentIdentifierFinished_Basic( self ):
    with MockCurrentFiletypes():
      with MockCurrentColumnAndLineContents( 3, 'ab;' ):
        assert_that( base.CurrentIdentifierFinished() )

      with MockCurrentColumnAndLineContents( 2, 'ab;' ):
        assert_that( not base.CurrentIdentifierFinished() )

      with MockCurrentColumnAndLineContents( 1, 'ab;' ):
        assert_that( not base.CurrentIdentifierFinished() )


  def test_CurrentIdentifierFinished_NothingBeforeColumn( self ):
    with MockCurrentColumnAndLineContents( 0, 'ab;' ):
      assert_that( base.CurrentIdentifierFinished() )

    with MockCurrentColumnAndLineContents( 0, '' ):
      assert_that( base.CurrentIdentifierFinished() )


  def test_CurrentIdentifierFinished_InvalidColumn( self ):
    with MockCurrentFiletypes():
      with MockCurrentColumnAndLineContents( 5, '' ):
        assert_that( base.CurrentIdentifierFinished() )

      with MockCurrentColumnAndLineContents( 5, 'abc' ):
        assert_that( not base.CurrentIdentifierFinished() )

      with MockCurrentColumnAndLineContents( 4, 'ab;' ):
        assert_that( base.CurrentIdentifierFinished() )


  def test_CurrentIdentifierFinished_InMiddleOfLine( self ):
    with MockCurrentFiletypes():
      with MockCurrentColumnAndLineContents( 4, 'bar.zoo' ):
        assert_that( base.CurrentIdentifierFinished() )

      with MockCurrentColumnAndLineContents( 4, 'bar(zoo' ):
        assert_that( base.CurrentIdentifierFinished() )

      with MockCurrentColumnAndLineContents( 4, 'bar-zoo' ):
        assert_that( base.CurrentIdentifierFinished() )


  def test_CurrentIdentifierFinished_Html( self ):
    with MockCurrentFiletypes( [ 'html' ] ):
      with MockCurrentColumnAndLineContents( 4, 'bar-zoo' ):
        assert_that( not base.CurrentIdentifierFinished() )


  def test_CurrentIdentifierFinished_WhitespaceOnly( self ):
    with MockCurrentFiletypes():
      with MockCurrentColumnAndLineContents( 1, '\n' ):
        assert_that( base.CurrentIdentifierFinished() )

      with MockCurrentColumnAndLineContents( 3, '\n    ' ):
        assert_that( base.CurrentIdentifierFinished() )

      with MockCurrentColumnAndLineContents( 3, '\t\t\t\t' ):
        assert_that( base.CurrentIdentifierFinished() )


  def test_CurrentIdentifierFinished_Unicode( self ):
    with MockCurrentFiletypes():
      # CurrentColumn returns a byte offset and character ø is 2 bytes length.
      with MockCurrentColumnAndLineContents( 6, 'føo ' ):
        assert_that( base.CurrentIdentifierFinished() )

      with MockCurrentColumnAndLineContents( 5, 'føo ' ):
        assert_that( base.CurrentIdentifierFinished() )

      with MockCurrentColumnAndLineContents( 4, 'føo ' ):
        assert_that( not base.CurrentIdentifierFinished() )

      with MockCurrentColumnAndLineContents( 3, 'føo ' ):
        assert_that( not base.CurrentIdentifierFinished() )



# Copyright (C) 2024 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, List
from ycm import vimsupport
import os


class HierarchyNode:
  def __init__( self, data, distance : int ):
    self._references : Optional[ List[ int ] ] = None
    self._data = data
    self._distance_from_root = distance


  def ToRootLocation( self, subindex : int ):
    location = self._data.get( 'root_location' )
    if location:
      file = location[ 'filepath' ]
      line = location[ 'line_num' ]
      column = location[ 'column_num' ]
      return file, line, column
    else:
      return self.ToLocation( subindex )


  def ToLocation( self, subindex : int ):
    location = self._data[ 'locations' ][ subindex ]
    line = location[ 'line_num' ]
    column = location[ 'column_num' ]
    file = location[ 'filepath' ]
    return file, line, column


MAX_HANDLES_PER_INDEX = 1000000


def handle_to_index( handle : int ):
  return abs( handle ) // MAX_HANDLES_PER_INDEX


def handle_to_location_index( handle : int ):
  return abs( handle ) % MAX_HANDLES_PER_INDEX


def make_handle( index : int, location_index : int ):
  return index * MAX_HANDLES_PER_INDEX + location_index


class HierarchyTree:
  def __init__( self ):
    self._up_nodes : List[ HierarchyNode ] = []
    self._down_nodes : List[ HierarchyNode ] = []
    self._kind : str = ''

  def SetRootNode( self, items, kind : str ):
    if items:
      assert len( items ) == 1
      self._root_node_indices = [ 0 ]
      self._down_nodes.append( HierarchyNode( items[ 0 ], 0 ) )
      self._up_nodes.append( HierarchyNode( items[ 0 ], 0 ) )
      self._kind = kind
      return self.HierarchyToLines()
    return []


  def UpdateHierarchy( self, handle : int, items, direction : str ):
    current_index = handle_to_index( handle )
    nodes = self._down_nodes if direction == 'down' else self._up_nodes
    if items:
      nodes.extend( [
        HierarchyNode( item,
                       nodes[ current_index ]._distance_from_root + 1 )
        for item in items ] )
      nodes[ current_index ]._references = list(
          range( len( nodes ) - len( items ), len( nodes ) ) )
    else:
      nodes[ current_index ]._references = []


  def Reset( self ):
    self._down_nodes = []
    self._up_nodes = []
    self._kind = ''


  def _HierarchyToLinesHelper( self, refs, use_down_nodes ):
    partial_result = []
    nodes = self._down_nodes if use_down_nodes else self._up_nodes
    for index in refs:
      next_node = nodes[ index ]
      indent = 2 * next_node._distance_from_root
      if index == 0:
        can_expand = ( self._down_nodes[ 0 ]._references is None or
                       self._up_nodes[ 0 ]._references is None )
      else:
        can_expand = next_node._references is None
      symbol = '+' if can_expand else '-'
      name = next_node._data[ 'name' ]
      kind = next_node._data[ 'kind' ]
      if use_down_nodes:
        partial_result.extend( [
          (
            {
              'indent': indent,
              'icon': symbol,
              'symbol': name,
              'kind': kind,
              'filepath': os.path.split( l[ 'filepath' ] )[ 1 ],
              'line_num': str( l[ 'line_num' ] ),
              'description': l.get( 'description', '' ).strip(),
            },
            make_handle( index, location_index )
          )
          for location_index, l in enumerate( next_node._data[ 'locations' ] )
        ] )
      else:
        partial_result.extend( [
          (
            {
              'indent': indent,
              'icon': symbol,
              'symbol': name,
              'kind': kind,
              'filepath': os.path.split( l[ 'filepath' ] )[ 1 ],
              'line_num': str( l[ 'line_num' ] ),
              'description': l.get( 'description', '' ).strip(),
            },
            make_handle( index, location_index ) * -1
          )
          for location_index, l in enumerate( next_node._data[ 'locations' ] )
        ] )
      if next_node._references:
        partial_result.extend(
          self._HierarchyToLinesHelper(
            next_node._references, use_down_nodes ) )
    return partial_result

  def HierarchyToLines( self ):
    down_lines = self._HierarchyToLinesHelper( [ 0 ], True )
    up_lines = self._HierarchyToLinesHelper( [ 0 ], False )
    up_lines.reverse()
    return up_lines + down_lines[ 1: ]


  def JumpToItem( self, handle : int, command ):
    node_index = handle_to_index( handle )
    location_index = handle_to_location_index( handle )
    if handle >= 0:
      node = self._down_nodes[ node_index ]
    else:
      node = self._up_nodes[ node_index ]
    file, line, column = node.ToLocation( location_index )
    vimsupport.JumpToLocation( file, line, column, '', command )


  def ShouldResolveItem( self, handle : int, direction : str ):
    node_index = handle_to_index( handle )
    if ( ( handle >= 0 and direction == 'down' ) or
         ( handle <= 0 and direction == 'up' ) ):
      if direction == 'down':
        node = self._down_nodes[ node_index ]
      else:
        node = self._up_nodes[ node_index ]
      return node._references is None
    return True


  def ResolveArguments( self, handle : int, direction : str ):
    node_index = handle_to_index( handle )
    if self._kind == 'call':
      direction = 'outgoing' if direction == 'up' else 'incoming'
    else:
      direction = 'supertypes' if direction == 'up' else 'subtypes'
    if handle >= 0:
      node = self._down_nodes[ node_index ]
    else:
      node = self._up_nodes[ node_index ]
    return [
      f'Resolve{ self._kind.title() }HierarchyItem',
      node._data,
      direction
    ]


  def HandleToRootLocation( self, handle : int ):
    node_index = handle_to_index( handle )

    if handle >= 0:
      node = self._down_nodes[ node_index ]
    else:
      node = self._up_nodes[ node_index ]

    location_index = handle_to_location_index( handle )
    return node.ToRootLocation( location_index )


  def UpdateChangesRoot( self, handle : int, direction : str ):
    return ( ( handle < 0 and direction == 'down' ) or
             ( handle > 0 and direction == 'up' ) )



# Copyright (C) 2020, YouCompleteMe Contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.


import logging
from ycm.client.base_request import ( BaseRequest, DisplayServerException,
                                      MakeServerException )

_logger = logging.getLogger( __name__ )


# FIXME: This is copy/pasta from SignatureHelpRequest - abstract a
# SimpleAsyncRequest base that does all of this generically
class SemanticTokensRequest( BaseRequest ):
  def __init__( self, request_data ):
    super().__init__()
    self.request_data = request_data
    self._response_future = None


  def Start( self ):
    self._response_future = self.PostDataToHandlerAsync( self.request_data,
                                                         'semantic_tokens' )

  def Done( self ):
    return bool( self._response_future ) and self._response_future.done()


  def Reset( self ):
    self._response_future = None

  def Response( self ):
    if not self._response_future:
      return {}

    response = self.HandleFuture( self._response_future,
                                  truncate_message = True )

    if not response:
      return {}

    # Vim may not be able to convert the 'errors' entry to its internal format
    # so we remove it from the response.
    errors = response.pop( 'errors', [] )
    for e in errors:
      exception = MakeServerException( e )
      _logger.error( exception )
      DisplayServerException( exception, truncate_message = True )

    return response.get( 'semantic_tokens' ) or {}



# encoding: utf-8
#
# Copyright (C) 2016-2019 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from hamcrest import assert_that, contains_exactly, empty, has_entries
from unittest import TestCase

from ycm.tests.test_utils import MockVimBuffers, MockVimModule, VimBuffer
MockVimModule()

from ycm import vimsupport
from ycm.tests import YouCompleteMeInstance, youcompleteme_instance

FILETYPE = 'ycmtest'
TRIGGERS = {
  'ycmtest': [ '.' ]
}


def StartColumnCompliance( ycm,
                           omnifunc_start_column,
                           ycm_completions,
                           ycm_start_column ):
  def Omnifunc( findstart, base ):
    if findstart:
      return omnifunc_start_column
    return [ 'foo' ]

  current_buffer = VimBuffer( 'buffer',
                              contents = [ 'fo' ],
                              filetype = FILETYPE,
                              omnifunc = Omnifunc )

  with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 2 ) ):
    ycm.SendCompletionRequest( force_semantic = True )
    r = ycm.GetCompletionResponse()
    assert_that(
      r,
      has_entries( {
        'completions': ycm_completions,
        'completion_start_column': ycm_start_column
      } )
    )


class OmniCompleterTest( TestCase ):
  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 1,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_Cache_List( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return [ 'a', 'b', 'cdef' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 5 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [
            { 'word': 'a',    'equal': 1 },
            { 'word': 'b',    'equal': 1 },
            { 'word': 'cdef', 'equal': 1 }
          ],
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 1,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_Cache_ListFilter( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return [ 'a', 'b', 'cdef' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.t' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 6 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': empty(),
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 0,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_NoCache_List( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return [ 'a', 'b', 'cdef' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 5 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [
            { 'word': 'a',    'equal': 1 },
            { 'word': 'b',    'equal': 1 },
            { 'word': 'cdef', 'equal': 1 }
          ],
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 0,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_NoCache_ListFilter( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return [ 'a', 'b', 'cdef' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.t' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 6 ) ):
      ycm.SendCompletionRequest()
      # Actual result is that the results are not filtered, as we expect the
      # omnifunc or vim itself to do this filtering.
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [
            { 'word': 'a',    'equal': 1 },
            { 'word': 'b',    'equal': 1 },
            { 'word': 'cdef', 'equal': 1 }
          ],
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 0,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_NoCache_UseFindStart( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 0
      return [ 'a', 'b', 'cdef' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.t' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 6 ) ):
      ycm.SendCompletionRequest()
      # Actual result is that the results are not filtered, as we expect the
      # omnifunc or vim itself to do this filtering.
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [
            { 'word': 'a',    'equal': 1 },
            { 'word': 'b',    'equal': 1 },
            { 'word': 'cdef', 'equal': 1 }
          ],
          'completion_start_column': 1
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 1,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_Cache_UseFindStart( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 0
      return [ 'a', 'b', 'cdef' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.t' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 6 ) ):
      ycm.SendCompletionRequest()
      # There are no results because the query 'test.t' doesn't match any
      # candidate (and cache_omnifunc=1, so we FilterAndSortCandidates).
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': empty(),
          'completion_start_column': 1
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 1,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_Cache_Object( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return { 'words': [ 'a', 'b', 'CDtEF' ] }

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.t' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 6 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [ { 'word': 'CDtEF', 'equal': 1 } ],
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 1,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_Cache_ObjectList( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return [
        {
          'word': 'a',
          'abbr': 'ABBR',
          'menu': 'MENU',
          'info': 'INFO',
          'kind': 'K'
        },
        {
          'word': 'test',
          'abbr': 'ABBRTEST',
          'menu': 'MENUTEST',
          'info': 'INFOTEST',
          'kind': 'T'
        }
      ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.tt' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 7 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': contains_exactly( {
            'word' : 'test',
            'abbr' : 'ABBRTEST',
            'menu' : 'MENUTEST',
            'info' : 'INFOTEST',
            'kind' : 'T',
            'equal': 1
          } ),
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 0,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_NoCache_ObjectList( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return [
        {
          'word': 'a',
          'abbr': 'ABBR',
          'menu': 'MENU',
          'info': 'INFO',
          'kind': 'K'
        },
        {
          'word': 'test',
          'abbr': 'ABBRTEST',
          'menu': 'MENUTEST',
          'info': 'INFOTEST',
          'kind': 'T'
        }
      ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.tt' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 7 ) ):
      ycm.SendCompletionRequest()
      # We don't filter the result - we expect the omnifunc to do that
      # based on the query we supplied (Note: that means no fuzzy matching!).
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [ {
            'word' : 'a',
            'abbr' : 'ABBR',
            'menu' : 'MENU',
            'info' : 'INFO',
            'kind' : 'K',
            'equal': 1
          }, {
            'word' : 'test',
            'abbr' : 'ABBRTEST',
            'menu' : 'MENUTEST',
            'info' : 'INFOTEST',
            'kind' : 'T',
            'equal': 1
          } ],
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 1,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_Cache_ObjectListObject( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return { 'words': [
        {
          'word': 'a',
          'abbr': 'ABBR',
          'menu': 'MENU',
          'info': 'INFO',
          'kind': 'K'
        },
        {
          'word': 'test',
          'abbr': 'ABBRTEST',
          'menu': 'MENUTEST',
          'info': 'INFOTEST',
          'kind': 'T'
        }
      ] }

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.tt' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 7 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [ {
            'word' : 'test',
            'abbr' : 'ABBRTEST',
            'menu' : 'MENUTEST',
            'info' : 'INFOTEST',
            'kind' : 'T',
            'equal': 1
          } ],
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 0,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_NoCache_ObjectListObject( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return { 'words': [
        {
          'word': 'a',
          'abbr': 'ABBR',
          'menu': 'MENU',
          'info': 'INFO',
          'kind': 'K'
        },
        {
          'word': 'test',
          'abbr': 'ABBRTEST',
          'menu': 'MENUTEST',
          'info': 'INFOTEST',
          'kind': 'T'
        }
      ] }

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.tt' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 7 ) ):
      ycm.SendCompletionRequest()
      # No FilterAndSortCandidates for cache_omnifunc=0 (we expect the omnifunc
      # to do the filtering?)
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [ {
            'word' : 'a',
            'abbr' : 'ABBR',
            'menu' : 'MENU',
            'info' : 'INFO',
            'kind' : 'K',
            'equal': 1
          }, {
            'word' : 'test',
            'abbr' : 'ABBRTEST',
            'menu' : 'MENUTEST',
            'info' : 'INFOTEST',
            'kind' : 'T',
            'equal': 1
          } ],
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 1,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_Cache_List_Unicode( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 12
      return [ '†est', 'å_unicode_identifier', 'πππππππ yummy πie' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ '†åsty_π.' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 12 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [
            { 'word': 'å_unicode_identifier', 'equal': 1 },
            { 'word': 'πππππππ yummy πie',    'equal': 1 },
            { 'word': '†est',                 'equal': 1 }
          ],
          'completion_start_column': 13
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 0,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_NoCache_List_Unicode( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 12
      return [ '†est', 'å_unicode_identifier', 'πππππππ yummy πie' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ '†åsty_π.' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 12 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [
            { 'word': '†est',                 'equal': 1 },
            { 'word': 'å_unicode_identifier', 'equal': 1 },
            { 'word': 'πππππππ yummy πie',    'equal': 1 }
          ],
          'completion_start_column': 13
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 1,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_Cache_List_Filter_Unicode( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 12
      return [ '†est', 'å_unicode_identifier', 'πππππππ yummy πie' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ '†åsty_π.ππ' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 17 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [ { 'word': 'πππππππ yummy πie', 'equal': 1 } ],
          'completion_start_column': 13
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 0,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_NoCache_List_Filter_Unicode(
      self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 12
      return [ 'πππππππ yummy πie' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ '†åsty_π.ππ' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 17 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [ { 'word': 'πππππππ yummy πie', 'equal': 1 } ],
          'completion_start_column': 13
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 1,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_Cache_ObjectList_Unicode( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 12
      return [
        {
          'word': 'ålpha∫et',
          'abbr': 'å∫∫®',
          'menu': 'µ´~¨á',
          'info': '^~fo',
          'kind': '˚'
        },
        {
          'word': 'π†´ß†π',
          'abbr': 'ÅııÂÊ‰ÍÊ',
          'menu': '˜‰ˆËÊ‰ÍÊ',
          'info': 'ÈˆÏØÊ‰ÍÊ',
          'kind': 'Ê'
        }
      ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ '†åsty_π.ππ' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 17 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [ {
            'word' : 'π†´ß†π',
            'abbr' : 'ÅııÂÊ‰ÍÊ',
            'menu' : '˜‰ˆËÊ‰ÍÊ',
            'info' : 'ÈˆÏØÊ‰ÍÊ',
            'kind' : 'Ê',
            'equal': 1
          } ],
          'completion_start_column': 13
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 1,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_Cache_ObjectListObject_Unicode(
      self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 12
      return {
        'words': [
          {
            'word': 'ålpha∫et',
            'abbr': 'å∫∫®',
            'menu': 'µ´~¨á',
            'info': '^~fo',
            'kind': '˚'
          },
          {
            'word': 'π†´ß†π',
            'abbr': 'ÅııÂÊ‰ÍÊ',
            'menu': '˜‰ˆËÊ‰ÍÊ',
            'info': 'ÈˆÏØÊ‰ÍÊ',
            'kind': 'Ê'
          },
          {
            'word': 'test',
            'abbr': 'ÅııÂÊ‰ÍÊ',
            'menu': '˜‰ˆËÊ‰ÍÊ',
            'info': 'ÈˆÏØÊ‰ÍÊ',
            'kind': 'Ê'
          }
        ]
      }

    current_buffer = VimBuffer( 'buffer',
                                contents = [ '†åsty_π.t' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 13 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': contains_exactly( {
            'word' : 'test',
            'abbr' : 'ÅııÂÊ‰ÍÊ',
            'menu' : '˜‰ˆËÊ‰ÍÊ',
            'info' : 'ÈˆÏØÊ‰ÍÊ',
            'kind' : 'Ê',
            'equal': 1
          }, {
            'word' : 'ålpha∫et',
            'abbr' : 'å∫∫®',
            'menu' : 'µ´~¨á',
            'info' : '^~fo',
            'kind' : '˚',
            'equal': 1
          } ),
          'completion_start_column': 13
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 1,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_RestoreCursorPositionAfterOmnifuncCall(
    self, ycm ):

    # This omnifunc moves the cursor to the test definition like
    # ccomplete#Complete would.
    def Omnifunc( findstart, base ):
      vimsupport.SetCurrentLineAndColumn( 0, 0 )
      if findstart:
        return 5
      return [ 'length' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'String test',
                                             '',
                                             'test.' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 3, 5 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        vimsupport.CurrentLineAndColumn(),
        contains_exactly( 2, 5 )
      )
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [ { 'word': 'length', 'equal': 1 } ],
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 1,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_MoveCursorPositionAtStartColumn(
      self, ycm ):
    # This omnifunc relies on the cursor being moved at the start column when
    # called the second time like LanguageClient#complete from the
    # LanguageClient-neovim plugin.
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      if vimsupport.CurrentColumn() == 5:
        return [ 'length' ]
      return []

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'String test',
                                             '',
                                             'test.le' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 3, 7 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        vimsupport.CurrentLineAndColumn(),
        contains_exactly( 2, 7 )
      )
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [ { 'word': 'length', 'equal': 1 } ],
          'completion_start_column': 6
        } )
      )


  def test_OmniCompleter_GetCompletions_StartColumnCompliance( self ):
    for omnifunc_start_column, ycm_completions, ycm_start_column in [
     [ -4, [ { 'word': 'foo', 'equal': 1 } ], 3 ],
     [ -3, [],                                1 ],
     [ -2, [],                                1 ],
     [ -1, [ { 'word': 'foo', 'equal': 1 } ], 3 ],
     [ 0, [ { 'word': 'foo', 'equal': 1 } ], 1 ],
     [ 1, [ { 'word': 'foo', 'equal': 1 } ], 2 ],
     [ 2, [ { 'word': 'foo', 'equal': 1 } ], 3 ],
     [ 3, [ { 'word': 'foo', 'equal': 1 } ], 3 ]
      ]:
      with youcompleteme_instance( { 'g:ycm_cache_omnifunc': 1 } ) as ycm:
        with self.subTest( omnifunc_start_column = omnifunc_start_column,
                           ycm_completions = ycm_completions,
                           ycm_start_column = ycm_start_column ):
          StartColumnCompliance( ycm,
                                 omnifunc_start_column,
                                 ycm_completions,
                                 ycm_start_column )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 0,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_NoCache_NoSemanticTrigger( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 0
      return [ 'test' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'te' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 3 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': empty(),
          'completion_start_column': 1
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 0,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_NoCache_ForceSemantic( self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 0
      return [ 'test' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'te' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 3 ) ):
      ycm.SendCompletionRequest( force_semantic = True )
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [ { 'word': 'test', 'equal': 1 } ],
          'completion_start_column': 1
        } )
      )


  @YouCompleteMeInstance( { 'g:ycm_cache_omnifunc': 1,
                            'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_ConvertStringsToDictionaries(
      self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return [
        { 'word': 'a' },
        'b'
      ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 7 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [
            { 'word': 'a', 'equal': 1 },
            { 'word': 'b', 'equal': 1 }
          ],
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( {
    'g:ycm_cache_omnifunc': 0,
    'g:ycm_filetype_specific_completion_to_disable': { FILETYPE: 1 },
    'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_FiletypeDisabled_SemanticTrigger(
      self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return [ 'a', 'b', 'cdef' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 6 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': empty(),
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( {
    'g:ycm_cache_omnifunc': 0,
    'g:ycm_filetype_specific_completion_to_disable': { '*': 1 },
    'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_AllFiletypesDisabled_SemanticTrigger(
    self, ycm ):

    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return [ 'a', 'b', 'cdef' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 6 ) ):
      ycm.SendCompletionRequest()
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': empty(),
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( {
    'g:ycm_cache_omnifunc': 0,
    'g:ycm_filetype_specific_completion_to_disable': { FILETYPE: 1 },
    'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_FiletypeDisabled_ForceSemantic(
      self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return [ 'a', 'b', 'cdef' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 6 ) ):
      ycm.SendCompletionRequest( force_semantic = True )
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [
            { 'word': 'a',    'equal': 1 },
            { 'word': 'b',    'equal': 1 },
            { 'word': 'cdef', 'equal': 1 }
          ],
          'completion_start_column': 6
        } )
      )


  @YouCompleteMeInstance( {
    'g:ycm_cache_omnifunc': 0,
    'g:ycm_filetype_specific_completion_to_disable': { '*': 1 },
    'g:ycm_semantic_triggers': TRIGGERS } )
  def test_OmniCompleter_GetCompletions_AllFiletypesDisabled_ForceSemantic(
      self, ycm ):
    def Omnifunc( findstart, base ):
      if findstart:
        return 5
      return [ 'a', 'b', 'cdef' ]

    current_buffer = VimBuffer( 'buffer',
                                contents = [ 'test.' ],
                                filetype = FILETYPE,
                                omnifunc = Omnifunc )

    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 6 ) ):
      ycm.SendCompletionRequest( force_semantic = True )
      assert_that(
        ycm.GetCompletionResponse(),
        has_entries( {
          'completions': [
            { 'word': 'a',    'equal': 1 },
            { 'word': 'b',    'equal': 1 },
            { 'word': 'cdef', 'equal': 1 }
          ],
          'completion_start_column': 6
        } )
      )



# Copyright (C) 2016-2018 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm.client.messages_request import MessagesPoll
from ycm.tests.test_utils import ( ExtendedMock,
                                   MockVimBuffers,
                                   MockVimModule,
                                   Version,
                                   VimBuffer,
                                   VimProp,
                                   VimSign )
MockVimModule()

import vim

import os
import sys
from hamcrest import ( assert_that, contains_exactly, empty, equal_to,
                       has_entries, is_in, is_not, matches_regexp )
from unittest.mock import call, MagicMock, patch
from unittest import TestCase

from ycm.paths import _PathToPythonUsedDuringBuild
from ycm.vimsupport import SetVariableValue
from ycm.tests import ( StopServer,
                        test_utils,
                        UserOptions,
                        WaitUntilReady,
                        YouCompleteMeInstance )
from ycm.client.base_request import _LoadExtraConfFile
from ycm.youcompleteme import YouCompleteMe
from ycmd.responses import ServerError
from ycm.tests.mock_utils import ( MockAsyncServerResponseDone,
                                   MockAsyncServerResponseInProgress,
                                   MockAsyncServerResponseException )


def RunNotifyUserIfServerCrashed( ycm, post_vim_message, test ):
  StopServer( ycm )

  ycm._logger = MagicMock( autospec = True )
  ycm._server_popen = MagicMock( autospec = True )
  ycm._server_popen.poll.return_value = test[ 'return_code' ]

  ycm.OnFileReadyToParse()

  assert_that( ycm._logger.error.call_args[ 0 ][ 0 ],
               test[ 'expected_message' ] )
  assert_that( post_vim_message.call_args[ 0 ][ 0 ],
               test[ 'expected_message' ] )


def YouCompleteMe_UpdateDiagnosticInterface( ycm, post_vim_message, *args ):

  contents = "\nint main() { int x, y; x == y }"

  # List of diagnostics returned by ycmd for the above code.
  diagnostics = [ {
    'kind': 'ERROR',
    'text': "expected ';' after expression (fix available) "
    "[expected_semi_after_expr]",
    'location': {
      'filepath': 'buffer',
      'line_num': 2,
      'column_num': 31
    },
    'location_extent': {
      'start': {
        'filepath': 'buffer',
        'line_num': 2,
        'column_num': 31,
      },
      'end': {
        'filepath': 'buffer',
        'line_num': 2,
        'column_num': 32,
      }
    },
    'ranges': [ {
      'start': {
        'line_num': 2,
        'column_num': 31,
        'filepath': 'buffer'
      },
      'end': {
        'line_num': 2,
        'column_num': 31,
        'filepath': 'buffer'
      }
    } ],
    'fixit_available': False
  }, {
    'kind': 'WARNING',
    'text': 'equality comparison result unused',
    'location': {
      'filepath': 'buffer',
      'line_num': 2,
      'column_num': 31,
    },
    'location_extent': {
      'start': {
        'filepath': 'buffer',
        'line_num': 2,
        'column_num': 31,
      },
      'end': {
        'filepath': 'buffer',
        'line_num': 2,
        'column_num': 32,
      }
    },
    'ranges': [ {
      'start': {
        'filepath': 'buffer',
        'line_num': 2,
        'column_num': 24,
      },
      'end': {
        'filepath': 'buffer',
        'line_num': 2,
        'column_num': 30,
      }
    } ],
    'fixit_available': False
  } ]

  current_buffer = VimBuffer( 'buffer',
                              filetype = 'c',
                              contents = contents.splitlines(),
                              number = 5 )

  test_utils.VIM_SIGNS = []
  with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 2, 1 ) ):
    with patch( 'ycm.client.event_notification.EventNotification.Response',
                return_value = diagnostics ):
      ycm.OnFileReadyToParse()
      ycm.HandleFileParseRequest( block = True )
    # The error on the current line is echoed, not the warning.
    post_vim_message.assert_called_once_with(
      "expected ';' after expression (fix available) "
      "[expected_semi_after_expr]",
      truncate = True, warning = False )

    # Error match is added after warning matches.
    assert_that(
      test_utils.VIM_PROPS_FOR_BUFFER,
      has_entries( {
        current_buffer.number: contains_exactly(
          VimProp( 'YcmWarningProperty', 2, 31, 2, 32 ),
          VimProp( 'YcmWarningProperty', 2, 24, 2, 30 ),
          VimProp( 'YcmErrorProperty', 2, 31, 2, 32 ),
          VimProp( 'YcmErrorProperty', 2, 31, 2, 31 ),
        )
      } )
    )

    # Only the error sign is placed.
    assert_that(
      test_utils.VIM_SIGNS,
      contains_exactly(
        VimSign( 2, 'YcmError', 5 )
      )
    )

  # The error is not echoed again when moving the cursor along the line.
  with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 2, 2 ) ):
    post_vim_message.reset_mock()
    ycm.OnCursorMoved()
    post_vim_message.assert_not_called()

  # The error is cleared when moving the cursor to another line.
  with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 2 ) ):
    post_vim_message.reset_mock()
    ycm.OnCursorMoved()
    post_vim_message.assert_called_once_with( "", warning = False )

  # The error is echoed when moving the cursor back.
  with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 2, 2 ) ):
    post_vim_message.reset_mock()
    ycm.OnCursorMoved()
    post_vim_message.assert_called_once_with(
      "expected ';' after expression (fix available) "
      "[expected_semi_after_expr]",
      truncate = True, warning = False )

    with patch( 'ycm.client.event_notification.EventNotification.Response',
                return_value = diagnostics[ 1 : ] ):
      ycm.OnFileReadyToParse()
      ycm.HandleFileParseRequest( block = True )

    assert_that(
      test_utils.VIM_PROPS_FOR_BUFFER,
      has_entries( {
        current_buffer.number: contains_exactly(
          VimProp( 'YcmWarningProperty', 2, 31, 2, 32 ),
          VimProp( 'YcmWarningProperty', 2, 24, 2, 30 )
        )
      } )
    )

    assert_that(
      test_utils.VIM_SIGNS,
      contains_exactly(
        VimSign( 2, 'YcmWarning', 5 )
      )
    )


class YouCompleteMeTest( TestCase ):
  @YouCompleteMeInstance()
  def test_YouCompleteMe_YcmCoreNotImported( self, ycm ):
    assert_that( 'ycm_core', is_not( is_in( sys.modules ) ) )


  @patch( 'ycm.vimsupport.PostVimMessage' )
  def test_YouCompleteMe_InvalidPythonInterpreterPath( self, post_vim_message ):
    with UserOptions( {
      'g:ycm_server_python_interpreter': '/invalid/python/path' } ):
      try:
        ycm = YouCompleteMe()

        assert_that( ycm.IsServerAlive(), equal_to( False ) )
        post_vim_message.assert_called_once_with(
          "Unable to start the ycmd server. "
          "Path in 'g:ycm_server_python_interpreter' option does not point "
          "to a valid Python 3.6+. "
          "Correct the error then restart the server with "
          "':YcmRestartServer'." )

        post_vim_message.reset_mock()

        SetVariableValue( 'g:ycm_server_python_interpreter',
                          _PathToPythonUsedDuringBuild() )
        ycm.RestartServer()

        assert_that( ycm.IsServerAlive(), equal_to( True ) )
        post_vim_message.assert_called_once_with( 'Restarting ycmd server...' )
      finally:
        WaitUntilReady()
        StopServer( ycm )


  @patch( 'ycmd.utils.PathToFirstExistingExecutable', return_value = None )
  @patch( 'ycm.paths._EndsWithPython', return_value = False )
  @patch( 'ycm.vimsupport.PostVimMessage' )
  def test_YouCompleteMe_NoPythonInterpreterFound(
      self, post_vim_message, *args ):
    with UserOptions( {} ):
      try:
        with patch( 'ycmd.utils.ReadFile', side_effect = IOError ):
          ycm = YouCompleteMe()

        assert_that( ycm.IsServerAlive(), equal_to( False ) )
        post_vim_message.assert_called_once_with(
          "Unable to start the ycmd server. Cannot find Python 3.6+. "
          "Set the 'g:ycm_server_python_interpreter' option to a Python "
          "interpreter path. "
          "Correct the error then restart the server with "
          "':YcmRestartServer'." )

        post_vim_message.reset_mock()

        SetVariableValue( 'g:ycm_server_python_interpreter',
                          _PathToPythonUsedDuringBuild() )
        ycm.RestartServer()

        assert_that( ycm.IsServerAlive(), equal_to( True ) )
        post_vim_message.assert_called_once_with( 'Restarting ycmd server...' )
      finally:
        WaitUntilReady()
        StopServer( ycm )


  @YouCompleteMeInstance()
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_YouCompleteMe_NotifyUserIfServerCrashed_UnexpectedCore(
      self, ycm, post_vim_message ):
    message = (
      "The ycmd server SHUT DOWN \\(restart with ':YcmRestartServer'\\). "
      "Unexpected error while loading the YCM core library. Type "
      "':YcmToggleLogs ycmd_\\d+_stderr_.+.log' to check the logs." )
    RunNotifyUserIfServerCrashed( ycm, post_vim_message, {
      'return_code': 3,
      'expected_message': matches_regexp( message )
    } )


  @YouCompleteMeInstance()
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_YouCompleteMe_NotifyUserIfServerCrashed_MissingCore(
      self, ycm, post_vim_message ):
    message = ( "The ycmd server SHUT DOWN (restart with ':YcmRestartServer'). "
                "YCM core library not detected; you need to compile YCM before "
                "using it. Follow the instructions in the documentation." )
    RunNotifyUserIfServerCrashed( ycm, post_vim_message, {
      'return_code': 4,
      'expected_message': equal_to( message )
    } )


  @YouCompleteMeInstance()
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_YouCompleteMe_NotifyUserIfServerCrashed_OutdatedCore(
      self, ycm, post_vim_message ):
    message = ( "The ycmd server SHUT DOWN (restart with ':YcmRestartServer'). "
                "YCM core library too old; PLEASE RECOMPILE by running the "
                "install.py script. See the documentation for more details." )
    RunNotifyUserIfServerCrashed( ycm, post_vim_message, {
      'return_code': 7,
      'expected_message': equal_to( message )
    } )


  @YouCompleteMeInstance()
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_YouCompleteMe_NotifyUserIfServerCrashed_UnexpectedExitCode(
      self, ycm, post_vim_message ):
    message = (
      "The ycmd server SHUT DOWN \\(restart with ':YcmRestartServer'\\). "
      "Unexpected exit code 1. Type "
      "':YcmToggleLogs ycmd_\\d+_stderr_.+.log' to check the logs." )
    RunNotifyUserIfServerCrashed( ycm, post_vim_message, {
      'return_code': 1,
      'expected_message': matches_regexp( message )
    } )


  @YouCompleteMeInstance( { 'g:ycm_extra_conf_vim_data': [ 'tempname()' ] } )
  @patch( 'ycm.vimsupport.VimSupportsPopupWindows', return_value=True )
  def test_YouCompleteMe_DebugInfo_ServerRunning( self, ycm, *args ):
    dir_of_script = os.path.dirname( os.path.abspath( __file__ ) )
    buf_name = os.path.join( dir_of_script, 'testdata', 'test.cpp' )
    extra_conf = os.path.join( dir_of_script, 'testdata', '.ycm_extra_conf.py' )
    _LoadExtraConfFile( extra_conf )

    current_buffer = VimBuffer( buf_name, filetype = 'cpp' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      assert_that(
        ycm.DebugInfo(),
        matches_regexp(
          'Client logfile: .+\n'
          'Server Python interpreter: .+\n'
          'Server Python version: .+\n'
          'Server has Clang support compiled in: (False|True)\n'
          'Clang version: .+\n'
          'Extra configuration file found and loaded\n'
          'Extra configuration path: .*testdata[/\\\\]\\.ycm_extra_conf\\.py\n'
          '[\\w\\W]*'
          'Server running at: .+\n'
          'Server process ID: \\d+\n'
          'Server logfiles:\n'
          '  .+\n'
          '  .+' )
      )


  @YouCompleteMeInstance()
  @patch( 'ycm.vimsupport.VimSupportsPopupWindows', return_value=True )
  def test_YouCompleteMe_DebugInfo_ServerNotRunning( self, ycm, *args ):
    StopServer( ycm )

    current_buffer = VimBuffer( 'current_buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      assert_that(
        ycm.DebugInfo(),
        matches_regexp(
          'Client logfile: .+\n'
          'Server errored, no debug info from server\n'
          'Server running at: .+\n'
          'Server process ID: \\d+\n'
          'Server logfiles:\n'
          '  .+\n'
          '  .+' )
      )


  @YouCompleteMeInstance()
  def test_YouCompleteMe_OnVimLeave_RemoveClientLogfileByDefault( self, ycm ):
    client_logfile = ycm._client_logfile
    assert_that( os.path.isfile( client_logfile ),
                 f'Logfile { client_logfile } does not exist.' )
    ycm.OnVimLeave()
    assert_that( not os.path.isfile( client_logfile ),
                 f'Logfile { client_logfile } was not removed.' )


  @YouCompleteMeInstance( { 'g:ycm_keep_logfiles': 1 } )
  def test_YouCompleteMe_OnVimLeave_KeepClientLogfile( self, ycm ):
    client_logfile = ycm._client_logfile
    assert_that( os.path.isfile( client_logfile ),
                 f'Logfile { client_logfile } does not exist.' )
    ycm.OnVimLeave()
    assert_that( os.path.isfile( client_logfile ),
                 f'Logfile { client_logfile } was removed.' )


  @YouCompleteMeInstance()
  @patch( 'ycm.vimsupport.CloseBuffersForFilename',
          new_callable = ExtendedMock )
  @patch( 'ycm.vimsupport.OpenFilename', new_callable = ExtendedMock )
  def test_YouCompleteMe_ToggleLogs_WithParameters(
      self, ycm, open_filename, close_buffers_for_filename ):
    logfile_buffer = VimBuffer( ycm._client_logfile )
    with MockVimBuffers( [ logfile_buffer ], [ logfile_buffer ] ):
      ycm.ToggleLogs( 90,
                      'botright vertical',
                      os.path.basename( ycm._client_logfile ),
                      'nonexisting_logfile',
                      os.path.basename( ycm._server_stdout ) )

      open_filename.assert_has_exact_calls( [
        call( ycm._server_stdout, { 'size': 90,
                                    'watch': True,
                                    'fix': True,
                                    'focus': False,
                                    'position': 'end',
                                    'mods': 'botright vertical' } )
      ] )
      close_buffers_for_filename.assert_has_exact_calls( [
        call( ycm._client_logfile )
      ] )


  @YouCompleteMeInstance()
  # Select the second item of the list which is the ycmd stderr logfile.
  @patch( 'ycm.vimsupport.SelectFromList', return_value = 1 )
  @patch( 'ycm.vimsupport.OpenFilename', new_callable = ExtendedMock )
  def test_YouCompleteMe_ToggleLogs_WithoutParameters_SelectLogfileNotAlreadyOpen( # noqa
    self, ycm, open_filename, *args ):

    current_buffer = VimBuffer( 'current_buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      ycm.ToggleLogs( 0, '' )

    open_filename.assert_has_exact_calls( [
      call( ycm._server_stderr, { 'size': 12,
                                  'watch': True,
                                  'fix': True,
                                  'focus': False,
                                  'position': 'end',
                                  'mods': '' } )
    ] )


  @YouCompleteMeInstance()
  # Select the third item of the list which is the ycmd stdout logfile.
  @patch( 'ycm.vimsupport.SelectFromList', return_value = 2 )
  @patch( 'ycm.vimsupport.CloseBuffersForFilename',
          new_callable = ExtendedMock )
  def test_YouCompleteMe_ToggleLogs_WithoutParameters_SelectLogfileAlreadyOpen(
    self, ycm, close_buffers_for_filename, *args ):

    logfile_buffer = VimBuffer( ycm._server_stdout )
    with MockVimBuffers( [ logfile_buffer ], [ logfile_buffer ] ):
      ycm.ToggleLogs( 0, '' )

    close_buffers_for_filename.assert_has_exact_calls( [
      call( ycm._server_stdout )
    ] )


  @YouCompleteMeInstance()
  @patch( 'ycm.vimsupport.SelectFromList',
          side_effect = RuntimeError( 'Error message' ) )
  @patch( 'ycm.vimsupport.PostVimMessage' )
  def test_YouCompleteMe_ToggleLogs_WithoutParameters_NoSelection(
    self, ycm, post_vim_message, *args ):

    current_buffer = VimBuffer( 'current_buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      ycm.ToggleLogs( 0, '' )

    assert_that(
      # Argument passed to PostVimMessage.
      post_vim_message.call_args[ 0 ][ 0 ],
      equal_to( 'Error message' )
    )


  @YouCompleteMeInstance()
  def test_YouCompleteMe_GetDefinedSubcommands_ListFromServer( self, ycm ):
    current_buffer = VimBuffer( 'buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with patch( 'ycm.client.base_request._JsonFromFuture',
                  return_value = [ 'SomeCommand', 'AnotherCommand' ] ):
        assert_that(
          ycm.GetDefinedSubcommands(),
          contains_exactly(
            'SomeCommand',
            'AnotherCommand'
          )
        )


  @YouCompleteMeInstance()
  @patch( 'ycm.client.base_request._logger', autospec = True )
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_YouCompleteMe_GetDefinedSubcommands_ErrorFromServer(
      self, ycm, post_vim_message, logger ):
    current_buffer = VimBuffer( 'buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with patch( 'ycm.client.base_request._JsonFromFuture',
                  side_effect = ServerError( 'Server error' ) ):
        result = ycm.GetDefinedSubcommands()

    logger.exception.assert_called_with(
        'Error while handling server response' )
    post_vim_message.assert_has_exact_calls( [
      call( 'Server error', truncate = False )
    ] )
    assert_that( result, empty() )


  @YouCompleteMeInstance()
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_YouCompleteMe_ShowDetailedDiagnostic_MessageFromServer(
    self, ycm, post_vim_message ):

    current_buffer = VimBuffer( 'buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with patch( 'ycm.client.base_request._JsonFromFuture',
                  return_value = { 'message': 'some_detailed_diagnostic' } ):
        ycm.ShowDetailedDiagnostic( False ),

    post_vim_message.assert_has_exact_calls( [
      call( 'some_detailed_diagnostic', warning = False )
    ] )


  @YouCompleteMeInstance()
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_YouCompleteMe_ShowDetailedDiagnostic_Exception(
    self, ycm, post_vim_message ):

    current_buffer = VimBuffer( 'buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with patch( 'ycm.client.base_request._JsonFromFuture',
                  side_effect = RuntimeError( 'Some exception' ) ):
        ycm.ShowDetailedDiagnostic( False ),

    post_vim_message.assert_has_exact_calls( [
      call( 'Some exception', truncate = False )
    ] )


  @YouCompleteMeInstance()
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_YouCompleteMe_ShowDiagnostics_FiletypeNotSupported(
    self, ycm, post_vim_message ):

    current_buffer = VimBuffer( 'buffer', filetype = 'not_supported' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      ycm.ShowDiagnostics()

    post_vim_message.assert_called_once_with(
      'Native filetype completion not supported for current file, '
      'cannot force recompilation.', warning = False )


  @YouCompleteMeInstance()
  @patch( 'ycm.youcompleteme.YouCompleteMe.FiletypeCompleterExistsForFiletype',
          return_value = True )
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  @patch( 'ycm.vimsupport.SetLocationListForWindow',
          new_callable = ExtendedMock )
  def test_YouCompleteMe_ShowDiagnostics_NoDiagnosticsDetected(
    self,
    ycm,
    set_location_list_for_window,
    post_vim_message,
    *args ):

    current_buffer = VimBuffer( 'buffer', filetype = 'cpp' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with patch( 'ycm.client.event_notification.EventNotification.Response',
                  return_value = {} ):
        ycm.ShowDiagnostics()
        set_location_list_for_window.assert_called_once_with(
          vim.current.window,
          [],
          1 )

    post_vim_message.assert_has_exact_calls( [
      call( 'Forcing compilation, this will block Vim until done.',
            warning = False ),
      call( 'Diagnostics refreshed', warning = False ),
      call( 'No warnings or errors detected.', warning = False )
    ] )


  @YouCompleteMeInstance( { 'g:ycm_log_level': 'debug',
                            'g:ycm_keep_logfiles': 1,
                            'g:ycm_open_loclist_on_ycm_diags': 0 } )
  @patch( 'ycm.youcompleteme.YouCompleteMe.FiletypeCompleterExistsForFiletype',
          return_value = True )
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  @patch( 'ycm.vimsupport.SetLocationListForWindow',
          new_callable = ExtendedMock )
  def test_YouCompleteMe_ShowDiagnostics_DiagnosticsFound_DoNotOpenLocationList(
    self,
    ycm,
    set_location_list_for_window,
    post_vim_message,
    *args ):
    diagnostic = {
      'kind': 'ERROR',
      'text': 'error text',
      'location': {
        'filepath': 'buffer',
        'line_num': 19,
        'column_num': 2
      },
      'location_extent': {
        'start': {
          'filepath': 'buffer',
          'line_num': 19,
          'column_num': 2
        },
        'end': {
          'filepath': 'buffer',
          'line_num': 19,
          'column_num': 3
        }
      }
    }

    current_buffer = VimBuffer( 'buffer',
                                filetype = 'cpp',
                                number = 3 )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with patch( 'ycm.client.event_notification.EventNotification.Response',
                  return_value = [ diagnostic ] ):
        ycm.ShowDiagnostics()
        set_location_list_for_window.assert_called_once_with(
          vim.current.window, [ {
            'bufnr': 3,
            'lnum': 19,
            'col': 2,
            'text': 'error text',
            'type': 'E',
            'valid': 1
          } ], 0 )

    post_vim_message.assert_has_exact_calls( [
      call( 'Forcing compilation, this will block Vim until done.',
            warning = False ),
      call( 'Diagnostics refreshed', warning = False )
    ] )


  @YouCompleteMeInstance( { 'g:ycm_open_loclist_on_ycm_diags': 1 } )
  @patch( 'ycm.youcompleteme.YouCompleteMe.FiletypeCompleterExistsForFiletype',
          return_value = True )
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  @patch( 'ycm.vimsupport.SetLocationListForWindow',
          new_callable = ExtendedMock )
  @patch( 'ycm.vimsupport.OpenLocationList', new_callable = ExtendedMock )
  def test_YouCompleteMe_ShowDiagnostics_DiagnosticsFound_OpenLocationList(
    self,
    ycm,
    open_location_list,
    set_location_list_for_window,
    post_vim_message,
    *args ):

    diagnostic = {
      'kind': 'ERROR',
      'text': 'error text',
      'location': {
        'filepath': 'buffer',
        'line_num': 19,
        'column_num': 2
      },
      'location_extent': {
        'start': {
          'filepath': 'buffer',
          'line_num': 19,
          'column_num': 2
        },
        'end': {
          'filepath': 'buffer',
          'line_num': 19,
          'column_num': 2
        }
      }
    }

    current_buffer = VimBuffer( 'buffer',
                                filetype = 'cpp',
                                number = 3 )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      with patch( 'ycm.client.event_notification.EventNotification.Response',
                  return_value = [ diagnostic ] ):
        ycm.ShowDiagnostics()
        set_location_list_for_window.assert_called_once_with(
          vim.current.window, [ {
            'bufnr': 3,
            'lnum': 19,
            'col': 2,
            'text': 'error text',
            'type': 'E',
            'valid': 1
          } ], 1 )

    post_vim_message.assert_has_exact_calls( [
      call( 'Forcing compilation, this will block Vim until done.',
            warning = False ),
      call( 'Diagnostics refreshed', warning = False )
    ] )
    open_location_list.assert_called_once_with( focus = True )


  @YouCompleteMeInstance( { 'g:ycm_echo_current_diagnostic': 1,
                            'g:ycm_enable_diagnostic_signs': 1,
                            'g:ycm_enable_diagnostic_highlighting': 1 } )
  @patch( 'ycm.youcompleteme.YouCompleteMe.FiletypeCompleterExistsForFiletype',
          return_value = True )
  @patch( 'ycm.client.event_notification.EventNotification.Done',
          return_value = True )
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_YouCompleteMe_UpdateDiagnosticInterface_OldVim(
    self, ycm, post_vim_message, *args ):
    YouCompleteMe_UpdateDiagnosticInterface( ycm, post_vim_message )


  @YouCompleteMeInstance( { 'g:ycm_echo_current_diagnostic': 1,
                            'g:ycm_enable_diagnostic_signs': 1,
                            'g:ycm_enable_diagnostic_highlighting': 1 } )
  @patch( 'ycm.youcompleteme.YouCompleteMe.FiletypeCompleterExistsForFiletype',
          return_value = True )
  @patch( 'ycm.tests.test_utils.VIM_VERSION', Version( 8, 1, 614 ) )
  @patch( 'ycm.client.event_notification.EventNotification.Done',
          return_value = True )
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_YouCompleteMe_UpdateDiagnosticInterface_NewVim(
    self, ycm, post_vim_message, *args ):
    YouCompleteMe_UpdateDiagnosticInterface( ycm, post_vim_message )


  @YouCompleteMeInstance( { 'g:ycm_enable_diagnostic_highlighting': 1 } )
  def test_YouCompleteMe_UpdateMatches_ClearDiagnosticMatchesInNewBuffer(
      self, ycm ):
    current_buffer = VimBuffer( 'buffer',
                                filetype = 'c',
                                contents = '\n\n\n\n',
                                number = 5 )

    test_utils.VIM_PROPS_FOR_BUFFER[ current_buffer.number ] = [
      VimProp( 'YcmWarningProperty', 3, 5, 3, 7 ),
      VimProp( 'YcmWarningProperty', 3, 3, 3, 9 ),
      VimProp( 'YcmErrorProperty', 3, 8 )
    ]

    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      ycm.UpdateMatches()

    assert_that( test_utils.VIM_PROPS_FOR_BUFFER[ current_buffer.number ],
                 empty() )


  @YouCompleteMeInstance( { 'g:ycm_echo_current_diagnostic': 1,
                            'g:ycm_always_populate_location_list': 1,
                            'g:ycm_show_diagnostics_ui': 0,
                            'g:ycm_enable_diagnostic_highlighting': 1 } )
  @patch( 'ycm.youcompleteme.YouCompleteMe.FiletypeCompleterExistsForFiletype',
          return_value = True )
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_YouCompleteMe_AsyncDiagnosticUpdate_UserDisabled(
      self,
      ycm,
      post_vim_message,
      *args ):
    diagnostics = [
      {
        'kind': 'ERROR',
        'text': 'error text in current buffer',
        'location': {
          'filepath': '/current',
          'line_num': 1,
          'column_num': 1
        },
        'location_extent': {
          'start': {
            'filepath': '/current',
            'line_num': 1,
            'column_num': 1,
          },
          'end': {
            'filepath': '/current',
            'line_num': 1,
            'column_num': 1,
          }
        },
        'ranges': []
      },
    ]
    current_buffer = VimBuffer( '/current',
                                filetype = 'ycmtest',
                                contents = [ 'current' ] * 10,
                                number = 1 )
    buffers = [ current_buffer ]
    windows = [ current_buffer ]

    # Register each buffer internally with YCM
    for current in buffers:
      with MockVimBuffers( buffers, [ current ] ):
        ycm.OnFileReadyToParse()
    with patch( 'ycm.vimsupport.SetLocationListForWindow',
                new_callable = ExtendedMock ) as set_location_list_for_window:
      with MockVimBuffers( buffers, windows ):
        ycm.UpdateWithNewDiagnosticsForFile( '/current', diagnostics )

    post_vim_message.assert_has_exact_calls( [] )
    set_location_list_for_window.assert_has_exact_calls( [] )

    assert_that(
      test_utils.VIM_PROPS_FOR_BUFFER,
      empty()
    )


  @YouCompleteMeInstance( { 'g:ycm_echo_current_diagnostic': 1,
                            'g:ycm_always_populate_location_list': 1,
                            'g:ycm_enable_diagnostic_highlighting': 1 } )
  @patch( 'ycm.youcompleteme.YouCompleteMe.FiletypeCompleterExistsForFiletype',
          return_value = True )
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_YouCompleteMe_AsyncDiagnosticUpdate_SingleFile(
      self,
      ycm,
      post_vim_message,
      *args ):

    # This test simulates asynchronous diagnostic updates associated with a
    # single file (e.g. Translation Unit), but where the actual errors refer to
    # other open files and other non-open files. This is not strictly invalid,
    # nor is it completely normal, but it is supported and does work.

    # Contrast with the next test which sends the diagnostics filewise, which
    # is what the language server protocol will do.

    diagnostics = [
      {
        'kind': 'ERROR',
        'text': 'error text in current buffer',
        'location': {
          'filepath': '/current',
          'line_num': 1,
          'column_num': 1
        },
        'location_extent': {
          'start': {
            'filepath': '/current',
            'line_num': 1,
            'column_num': 1,
          },
          'end': {
            'filepath': '/current',
            'line_num': 1,
            'column_num': 1,
          }
        },
        'ranges': []
      },
      {
        'kind': 'ERROR',
        'text': 'error text in hidden buffer',
        'location': {
          'filepath': '/has_diags',
          'line_num': 4,
          'column_num': 2
        },
        'location_extent': {
          'start': {
            'filepath': '/has_diags',
            'line_num': 4,
            'column_num': 2,
          },
          'end': {
            'filepath': '/has_diags',
            'line_num': 4,
            'column_num': 2,
          }
        },
        'ranges': []
      },
      {
        'kind': 'ERROR',
        'text': 'error text in buffer not open in Vim',
        'location': {
          'filepath': '/not_open',
          'line_num': 8,
          'column_num': 4
        },
        'location_extent': {
          'start': {
            'filepath': '/not_open',
            'line_num': 8,
            'column_num': 4,
          },
          'end': {
            'filepath': '/not_open',
            'line_num': 8,
            'column_num': 4,
          }
        },
        'ranges': []
      }
    ]

    current_buffer = VimBuffer( '/current',
                                filetype = 'ycmtest',
                                contents = [ 'current' ] * 10,
                                number = 1 )
    no_diags_buffer = VimBuffer( '/no_diags',
                                 filetype = 'ycmtest',
                                 contents = [ 'nodiags' ] * 10,
                                 number = 2 )
    hidden_buffer = VimBuffer( '/has_diags',
                               filetype = 'ycmtest',
                               contents = [ 'hasdiags' ] * 10,
                               number = 3 )

    buffers = [ current_buffer, no_diags_buffer, hidden_buffer ]
    windows = [ current_buffer, no_diags_buffer ]

    # Register each buffer internally with YCM
    for current in buffers:
      with MockVimBuffers( buffers, [ current ] ):
        ycm.OnFileReadyToParse()

    with patch( 'ycm.vimsupport.SetLocationListForWindow',
                new_callable = ExtendedMock ) as set_location_list_for_window:
      with MockVimBuffers( buffers, windows ):
        ycm.UpdateWithNewDiagnosticsForFile( '/current', diagnostics )
        # Ensure we included all the diags though
        set_location_list_for_window.assert_has_exact_calls( [
          call( vim.current.window, [
            {
              'lnum': 1,
              'col': 1,
              'bufnr': 1,
              'valid': 1,
              'type': 'E',
              'text': 'error text in current buffer',
            },
            {
              'lnum': 4,
              'col': 2,
              'bufnr': 3,
              'valid': 1,
              'type': 'E',
              'text': 'error text in hidden buffer',
            },
            {
              'lnum': 8,
              'col': 4,
              'bufnr': -1, # sic: Our mocked bufnr function actually returns -1,
                           # even though YCM is passing "create if needed".
                           # FIXME? we shouldn't do that, and we should pass
                           # filename instead
              'valid': 1,
              'type': 'E',
              'text': 'error text in buffer not open in Vim'
            }
          ], False )
        ] )


    # We update the diagnostic on the current cursor position
    post_vim_message.assert_has_exact_calls( [
      call( "error text in current buffer", truncate = True, warning = False ),
    ] )

    assert_that(
      test_utils.VIM_PROPS_FOR_BUFFER,
      has_entries( {
        1: contains_exactly(
          VimProp( 'YcmErrorProperty', 1, 1, 1, 1 )
        )
      } )
    )


  @YouCompleteMeInstance( { 'g:ycm_echo_current_diagnostic': 1,
                            'g:ycm_always_populate_location_list': 1,
                            'g:ycm_enable_diagnostic_highlighting': 1 } )
  @patch( 'ycm.youcompleteme.YouCompleteMe.FiletypeCompleterExistsForFiletype',
          return_value = True )
  @patch( 'ycm.vimsupport.PostVimMessage', new_callable = ExtendedMock )
  def test_YouCompleteMe_AsyncDiagnosticUpdate_PerFile(
      self,
      ycm,
      post_vim_message,
      *args ):

    # This test simulates asynchronous diagnostic updates which are delivered
    # per file, including files which are open and files which are not.

    # Ordered to ensure that the calls to update are deterministic
    diagnostics_per_file = [
      ( '/current', [ {
          'kind': 'ERROR',
          'text': 'error text in current buffer',
          'location': {
            'filepath': '/current',
            'line_num': 1,
            'column_num': 1
          },
          'location_extent': {
            'start': {
              'filepath': '/current',
              'line_num': 1,
              'column_num': 1,
            },
            'end': {
              'filepath': '/current',
              'line_num': 1,
              'column_num': 1,
            }
          },
          'ranges': [],
        } ] ),
      ( '/separate_window', [ {
          'kind': 'ERROR',
          'text': 'error text in a buffer open in a separate window',
          'location': {
            'filepath': '/separate_window',
            'line_num': 3,
            'column_num': 3
          },
          'location_extent': {
            'start': {
              'filepath': '/separate_window',
              'line_num': 3,
              'column_num': 3,
            },
            'end': {
              'filepath': '/separate_window',
              'line_num': 3,
              'column_num': 3,
            }
          },
          'ranges': []
        } ] ),
      ( '/hidden', [ {
          'kind': 'ERROR',
          'text': 'error text in hidden buffer',
          'location': {
            'filepath': '/hidden',
            'line_num': 4,
            'column_num': 2
          },
          'location_extent': {
            'start': {
              'filepath': '/hidden',
              'line_num': 4,
              'column_num': 2,
            },
            'end': {
              'filepath': '/hidden',
              'line_num': 4,
              'column_num': 2,
            }
          },
          'ranges': []
        } ] ),
      ( '/not_open', [ {
          'kind': 'ERROR',
          'text': 'error text in buffer not open in Vim',
          'location': {
            'filepath': '/not_open',
            'line_num': 8,
            'column_num': 4
          },
          'location_extent': {
            'start': {
              'filepath': '/not_open',
              'line_num': 8,
              'column_num': 4,
            },
            'end': {
              'filepath': '/not_open',
              'line_num': 8,
              'column_num': 4,
            }
          },
          'ranges': []
        } ] )
    ]

    current_buffer = VimBuffer( '/current',
                                filetype = 'ycmtest',
                                contents = [ 'current' ] * 10,
                                number = 1 )
    no_diags_buffer = VimBuffer( '/no_diags',
                                 filetype = 'ycmtest',
                                 contents = [ 'no_diags' ] * 10,
                                 number = 2 )
    separate_window = VimBuffer( '/separate_window',
                                 filetype = 'ycmtest',
                                 contents = [ 'separate_window' ] * 10,
                                 number = 3 )
    hidden_buffer = VimBuffer( '/hidden',
                               filetype = 'ycmtest',
                               contents = [ 'hidden' ] * 10,
                               number = 4 )
    buffers = [
      current_buffer,
      no_diags_buffer,
      separate_window,
      hidden_buffer
    ]
    windows = [
      current_buffer,
      no_diags_buffer,
      separate_window
    ]

    # Register each buffer internally with YCM
    for current in buffers:
      with MockVimBuffers( buffers, [ current ] ):
        ycm.OnFileReadyToParse()

    with patch( 'ycm.vimsupport.SetLocationListForWindow',
                new_callable = ExtendedMock ) as set_location_list_for_window:
      with MockVimBuffers( buffers, windows ):
        for filename, diagnostics in diagnostics_per_file:
          ycm.UpdateWithNewDiagnosticsForFile( filename, diagnostics )

        # Ensure we included all the diags though
        set_location_list_for_window.assert_has_exact_calls( [
          call( vim.windows[ 0 ], [
            {
              'lnum': 1,
              'col': 1,
              'bufnr': 1,
              'valid': 1,
              'type': 'E',
              'text': 'error text in current buffer',
            },
          ], False ),

          call( vim.windows[ 2 ], [
            {
              'lnum': 3,
              'col': 3,
              'bufnr': 3,
              'valid': 1,
              'type': 'E',
              'text': 'error text in a buffer open in a separate window',
            },
          ], False )
        ] )


    # We update the diagnostic on the current cursor position
    post_vim_message.assert_has_exact_calls( [
      call( "error text in current buffer", truncate = True, warning = False ),
    ] )

    assert_that(
      test_utils.VIM_PROPS_FOR_BUFFER,
      has_entries( {
        1: contains_exactly(
          VimProp( 'YcmErrorProperty', 1, 1, 1, 1 )
        ),
        3: contains_exactly(
          VimProp( 'YcmErrorProperty', 3, 3, 3, 3 )
        )
      } )
    )


  @YouCompleteMeInstance()
  def test_YouCompleteMe_OnPeriodicTick_ServerNotRunning( self, ycm ):
    with patch.object( ycm, 'IsServerAlive', return_value = False ):
      assert_that( ycm.OnPeriodicTick(), equal_to( False ) )


  @YouCompleteMeInstance()
  def test_YouCompleteMe_OnPeriodicTick_ServerNotReady( self, ycm ):
    with patch.object( ycm, 'IsServerAlive', return_value = True ):
      with patch.object( ycm, 'IsServerReady', return_value = False ):
        assert_that( ycm.OnPeriodicTick(), equal_to( True ) )


  @YouCompleteMeInstance()
  @patch( 'ycm.youcompleteme.YouCompleteMe.FiletypeCompleterExistsForFiletype',
          return_value = True )
  @patch( 'ycm.client.base_request._ValidateResponseObject',
          return_value = True )
  @patch( 'ycm.client.base_request.BaseRequest.PostDataToHandlerAsync' )
  def test_YouCompleteMe_OnPeriodicTick_DontRetry(
      self,
      ycm,
      post_data_to_handler_async,
      *args ):

    current_buffer = VimBuffer( '/current',
                                filetype = 'ycmtest',
                                number = 1 )

    # Create the request and make the first poll; we expect no response
    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 1 ) ):
      assert_that( ycm.OnPeriodicTick(), equal_to( True ) )
      post_data_to_handler_async.assert_called()

    assert ycm._message_poll_requests[ 'ycmtest' ] is not None
    post_data_to_handler_async.reset_mock()

    # OK that sent the request, now poll to check if it is complete (say it is
    # not)
    with MockVimBuffers( [ current_buffer ],
                         [ current_buffer ],
                         ( 1, 1 ) ) as v:
      mock_response = MockAsyncServerResponseInProgress()
      with patch.dict( ycm._message_poll_requests, {} ):
        ycm._message_poll_requests[ 'ycmtest' ] = MessagesPoll(
                                                    v.current.buffer )
        ycm._message_poll_requests[ 'ycmtest' ]._response_future = mock_response
        mock_future = ycm._message_poll_requests[ 'ycmtest' ]._response_future
        poll_again = ycm.OnPeriodicTick()
        mock_future.done.assert_called()
        mock_future.result.assert_not_called()
        assert_that( poll_again, equal_to( True ) )

    # Poll again, but return a response (telling us to stop polling)
    with MockVimBuffers( [ current_buffer ],
                         [ current_buffer ],
                         ( 1, 1 ) ) as v:
      mock_response = MockAsyncServerResponseDone( False )
      with patch.dict( ycm._message_poll_requests, {} ):
        ycm._message_poll_requests[ 'ycmtest' ] = MessagesPoll(
                                                    v.current.buffer )
        ycm._message_poll_requests[ 'ycmtest' ]._response_future = mock_response
        mock_future = ycm._message_poll_requests[ 'ycmtest' ]._response_future
        poll_again = ycm.OnPeriodicTick()
        mock_future.done.assert_called()
        mock_future.result.assert_called()
        post_data_to_handler_async.assert_not_called()
        # We reset and don't poll anymore
        assert_that( ycm._message_poll_requests[ 'ycmtest' ] is None )
        assert_that( poll_again, equal_to( False ) )



  @YouCompleteMeInstance()
  @patch( 'ycm.youcompleteme.YouCompleteMe.FiletypeCompleterExistsForFiletype',
          return_value = True )
  @patch( 'ycm.client.base_request._ValidateResponseObject',
          return_value = True )
  @patch( 'ycm.client.base_request.BaseRequest.PostDataToHandlerAsync' )
  def test_YouCompleteMe_OnPeriodicTick_Exception(
      self,
      ycm,
      post_data_to_handler_async,
      *args ):

    current_buffer = VimBuffer( '/current',
                                filetype = 'ycmtest',
                                number = 1 )

    # Create the request and make the first poll; we expect no response
    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 1 ) ):
      assert_that( ycm.OnPeriodicTick(), equal_to( True ) )
      post_data_to_handler_async.assert_called()

    post_data_to_handler_async.reset_mock()

    # Poll again, but return an exception response
    with MockVimBuffers( [ current_buffer ],
                         [ current_buffer ],
                         ( 1, 1 ) ) as v:
      mock_response = MockAsyncServerResponseException(
                        RuntimeError( 'test' ) )
      with patch.dict( ycm._message_poll_requests, {} ):
        ycm._message_poll_requests[ 'ycmtest' ] = MessagesPoll(
                                                    v.current.buffer )
        ycm._message_poll_requests[ 'ycmtest' ]._response_future = mock_response
        mock_future = ycm._message_poll_requests[ 'ycmtest' ]._response_future
        assert_that( ycm.OnPeriodicTick(), equal_to( False ) )
        mock_future.done.assert_called()
        mock_future.result.assert_called()
        post_data_to_handler_async.assert_not_called()
        # We reset and don't poll anymore
        assert_that( ycm._message_poll_requests[ 'ycmtest' ] is None )


  @YouCompleteMeInstance()
  @patch( 'ycm.youcompleteme.YouCompleteMe.FiletypeCompleterExistsForFiletype',
          return_value = True )
  @patch( 'ycm.client.base_request._ValidateResponseObject',
          return_value = True )
  @patch( 'ycm.client.base_request.BaseRequest.PostDataToHandlerAsync' )
  @patch( 'ycm.client.messages_request._HandlePollResponse' )
  def test_YouCompleteMe_OnPeriodicTick_ValidResponse(
      self, ycm, handle_poll_response, post_data_to_handler_async, *args ):

    current_buffer = VimBuffer( '/current',
                                filetype = 'ycmtest',
                                number = 1 )

    # Create the request and make the first poll; we expect no response
    with MockVimBuffers( [ current_buffer ],
                         [ current_buffer ],
                         ( 1, 1 ) ):
      assert_that( ycm.OnPeriodicTick(), equal_to( True ) )
      post_data_to_handler_async.assert_called()

    post_data_to_handler_async.reset_mock()

    # Poll again, and return a _proper_ response (finally!).
    # Note, _HandlePollResponse is tested independently (for simplicity)
    with MockVimBuffers( [ current_buffer ],
                         [ current_buffer ],
                         ( 1, 1 ) ) as v:
      mock_response = MockAsyncServerResponseDone( [] )
      with patch.dict( ycm._message_poll_requests, {} ):
        ycm._message_poll_requests[ 'ycmtest' ] = MessagesPoll(
                                                    v.current.buffer )
        ycm._message_poll_requests[ 'ycmtest' ]._response_future = mock_response
        mock_future = ycm._message_poll_requests[ 'ycmtest' ]._response_future
        assert_that( ycm.OnPeriodicTick(), equal_to( True ) )
        handle_poll_response.assert_called()
        mock_future.done.assert_called()
        mock_future.result.assert_called()
        post_data_to_handler_async.assert_called() # Poll again!
        assert_that( ycm._message_poll_requests[ 'ycmtest' ] is not None )


  @YouCompleteMeInstance()
  @patch( 'ycm.client.completion_request.CompletionRequest.OnCompleteDone' )
  def test_YouCompleteMe_OnCompleteDone_CompletionRequest(
      self, ycm, on_complete_done ):
    current_buffer = VimBuffer( 'current_buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ], ( 1, 1 ) ):
      ycm.SendCompletionRequest()
    ycm.OnCompleteDone()
    on_complete_done.assert_called()


  @YouCompleteMeInstance()
  @patch( 'ycm.client.completion_request.CompletionRequest.OnCompleteDone' )
  def test_YouCompleteMe_OnCompleteDone_NoCompletionRequest(
      self, ycm, on_complete_done ):
    ycm.OnCompleteDone()
    on_complete_done.assert_not_called()


  @YouCompleteMeInstance()
  def test_YouCompleteMe_ShouldResendFileParseRequest_NoParseRequest(
      self, ycm ):
    current_buffer = VimBuffer( 'current_buffer' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      assert_that( ycm.ShouldResendFileParseRequest(), equal_to( False ) )



# Copyright (C) 2011-2019 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict, namedtuple
from unittest.mock import DEFAULT, MagicMock, patch
from unittest import skip
from hamcrest import ( assert_that,
                       contains_exactly,
                       contains_inanyorder,
                       equal_to )
import contextlib
import functools
import json
import os
import re
import sys

from unittest import skipIf

from ycmd.utils import GetCurrentDirectory, OnMac, OnWindows, ToUnicode


BUFNR_REGEX = re.compile(
  '^bufnr\\( \'(?P<buffer_filename>.+)\'(, ([01]))? \\)$' )
BUFWINNR_REGEX = re.compile( '^bufwinnr\\( (?P<buffer_number>[0-9]+) \\)$' )
BWIPEOUT_REGEX = re.compile(
  '^(?:silent! )bwipeout!? (?P<buffer_number>[0-9]+)$' )
GETBUFVAR_REGEX = re.compile(
  '^getbufvar\\((?P<buffer_number>[0-9]+), "(?P<option>.+)"\\)( \\?\\? 0)?$' )
PROP_ADD_REGEX = re.compile(
        '^prop_add\\( '            # A literal at the start
        '(?P<start_line>\\d+), '   # First argument - number
        '(?P<start_column>\\d+), ' # Second argument - number
        '{(?P<opts>.+)} '          # Third argument is a complex dict, which
                                   # we parse separately
        '\\)$' )
PROP_REMOVE_REGEX = re.compile( '^prop_remove\\( (?P<prop>.+) \\)$' )
OMNIFUNC_REGEX_FORMAT = (
  '^{omnifunc_name}\\((?P<findstart>[01]),[\'"](?P<base>.*)[\'"]\\)$' )
FNAMEESCAPE_REGEX = re.compile( '^fnameescape\\(\'(?P<filepath>.+)\'\\)$' )
STRDISPLAYWIDTH_REGEX = re.compile(
  '^strdisplaywidth\\( ?\'(?P<text>.+)\' ?\\)$' )
REDIR_START_REGEX = re.compile( '^redir => (?P<variable>[\\w:]+)$' )
REDIR_END_REGEX = re.compile( '^redir END$' )
EXISTS_REGEX = re.compile( '^exists\\( \'(?P<option>[\\w:]+)\' \\)$' )
LET_REGEX = re.compile( '^let (?P<option>[\\w:]+) = (?P<value>.*)$' )
HAS_PATCH_REGEX = re.compile( '^has\\( \'patch(?P<patch>\\d+)\' \\)$' )

# One-and only instance of mocked Vim object. The first 'import vim' that is
# executed binds the vim module to the instance of MagicMock that is created,
# and subsquent assignments to sys.modules[ 'vim' ] don't retrospectively
# update them. The result is that while running the tests, we must assign only
# one instance of MagicMock to sys.modules[ 'vim' ] and always return it.
#
# More explanation is available:
# https://github.com/Valloric/YouCompleteMe/pull/1694
VIM_MOCK = MagicMock()

VIM_PROPS_FOR_BUFFER = defaultdict( list )
VIM_SIGNS = []

VIM_OPTIONS = {
  '&completeopt': b'',
  '&previewheight': 12,
  '&columns': 80,
  '&ruler': 0,
  '&showcmd': 1,
  '&hidden': 0,
  '&expandtab': 1
}

Version = namedtuple( 'Version', [ 'major', 'minor', 'patch' ] )

# This variable must be patched with a Version object for tests depending on a
# recent Vim version. Example:
#
#   @patch( 'ycm.tests.test_utils.VIM_VERSION', Version( 8, 1, 614 ) )
#   def ThisTestDependsOnTheVimVersion_test():
#     ...
#
# Default is the oldest supported version.
VIM_VERSION = Version( 7, 4, 1578 )

REDIR = {
  'status': False,
  'variable': '',
  'output': ''
}

WindowsAndMacOnly = skipIf( not OnWindows() or not OnMac(),
                            'Windows and macOS only' )


@contextlib.contextmanager
def CurrentWorkingDirectory( path ):
  old_cwd = GetCurrentDirectory()
  os.chdir( path )
  try:
    yield
  finally:
    os.chdir( old_cwd )


def _MockGetBufferNumber( buffer_filename ):
  for vim_buffer in VIM_MOCK.buffers:
    if vim_buffer.name == buffer_filename:
      return vim_buffer.number
  return -1


def _MockGetBufferWindowNumber( buffer_number ):
  for window in VIM_MOCK.windows:
    if window.buffer.number == buffer_number:
      return window.number
  return -1


def _MockGetBufferVariable( buffer_number, option ):
  for vim_buffer in VIM_MOCK.buffers:
    if vim_buffer.number == buffer_number:
      if option == '&mod':
        return vim_buffer.modified
      if option == '&ft':
        return vim_buffer.filetype
      if option == 'changedtick':
        return vim_buffer.changedtick
      if option == '&bh':
        return vim_buffer.bufhidden
      return ''
  return ''


def _MockVimBufferEval( value ):
  if value == '&omnifunc':
    return VIM_MOCK.current.buffer.omnifunc_name

  if value == '&filetype':
    return VIM_MOCK.current.buffer.filetype

  match = BUFNR_REGEX.search( value )
  if match:
    buffer_filename = match.group( 'buffer_filename' )
    return _MockGetBufferNumber( buffer_filename )

  match = BUFWINNR_REGEX.search( value )
  if match:
    buffer_number = int( match.group( 'buffer_number' ) )
    return _MockGetBufferWindowNumber( buffer_number )

  match = GETBUFVAR_REGEX.search( value )
  if match:
    buffer_number = int( match.group( 'buffer_number' ) )
    option = match.group( 'option' )
    return _MockGetBufferVariable( buffer_number, option )

  current_buffer = VIM_MOCK.current.buffer
  match = re.search( OMNIFUNC_REGEX_FORMAT.format(
                         omnifunc_name = current_buffer.omnifunc_name ),
                     value )
  if match:
    findstart = int( match.group( 'findstart' ) )
    base = match.group( 'base' )
    return current_buffer.omnifunc( findstart, base )

  return None


def _MockVimWindowEval( value ):
  if value == 'winnr("#")':
    # For simplicity, we always assume there is no previous window.
    return 0

  return None


def _MockVimOptionsEval( value ):
  result = VIM_OPTIONS.get( value )
  if result is not None:
    return result

  if value == 'keys( g: )':
    global_options = {}
    for key, value in VIM_OPTIONS.items():
      if key.startswith( 'g:' ):
        global_options[ key[ 2: ] ] = value
    return global_options

  match = EXISTS_REGEX.search( value )
  if match:
    option = match.group( 'option' )
    return option in VIM_OPTIONS

  return None


def _MockVimFunctionsEval( value ):
  if value == 'tempname()':
    return '_TEMP_FILE_'

  if value == 'tagfiles()':
    return [ 'tags' ]

  if value == 'shiftwidth()':
    return 2

  if value.startswith( 'has( "' ):
    return False

  match = re.match( 'sign_getplaced\\( (?P<bufnr>\\d+), '
                    '{ "group": "ycm_signs" } \\)', value )
  if match:
    filtered = list( filter( lambda sign: sign.bufnr ==
                                          int( match.group( 'bufnr' ) ),
                             VIM_SIGNS ) )
    r = [ { 'signs': filtered } ]
    return r

  match = re.match( 'sign_unplacelist\\( (?P<sign_list>\\[.*\\]) \\)', value )
  if match:
    sign_list = eval( match.group( 'sign_list' ) )
    for sign in sign_list:
      VIM_SIGNS.remove( sign )
    return True # Why True?

  match = re.match( 'sign_placelist\\( (?P<sign_list>\\[.*\\]) \\)', value )
  if match:
    sign_list = json.loads( match.group( 'sign_list' ).replace( "'", '"' ) )
    for sign in sign_list:
      VIM_SIGNS.append( VimSign( sign[ 'lnum' ],
                                 sign[ 'name' ],
                                 sign[ 'buffer' ] ) )
    return True # Why True?

  return None


def _MockVimPropEval( value ):
  match = re.match( 'prop_list\\( (?P<lnum>\\d+), '
                    '{ "bufnr": (?P<bufnr>\\d+) } \\)', value )
  if match:
    return [ p for p in VIM_PROPS_FOR_BUFFER[ int( match.group( 'bufnr' ) ) ]
             if p.start_line == int( match.group( 'lnum' ) ) ]

  match = PROP_ADD_REGEX.search( value )
  if match:
    prop_start_line = int( match.group( 'start_line' ) )
    prop_start_column = int( match.group( 'start_column' ) )
    import ast
    opts = ast.literal_eval( '{' + match.group( 'opts' ) + '}' )
    vim_prop = VimProp(
        opts[ 'type' ],
        prop_start_line,
        prop_start_column,
        int( opts[ 'end_lnum' ] ) if opts[ 'end_lnum' ] else prop_start_line,
        int( opts[ 'end_col' ] ) if opts[ 'end_col' ] else prop_start_column
    )
    VIM_PROPS_FOR_BUFFER[ int( opts[ 'bufnr' ] ) ].append( vim_prop )
    return vim_prop.id

  match = PROP_REMOVE_REGEX.search( value )
  if match:
    prop, lin_num = eval( match.group( 'prop' ) )
    vim_props = VIM_PROPS_FOR_BUFFER[ prop[ 'bufnr' ] ]
    for index, vim_prop in enumerate( vim_props ):
      if vim_prop.id == prop[ 'id' ]:
        vim_props.pop( index )
        return -1
    return 0

  return None


def _MockVimVersionEval( value ):
  match = HAS_PATCH_REGEX.search( value )
  if match:
    if not isinstance( VIM_VERSION, Version ):
      raise RuntimeError( 'Vim version is not set.' )
    return VIM_VERSION.patch >= int( match.group( 'patch' ) )

  if value == 'v:version':
    if not isinstance( VIM_VERSION, Version ):
      raise RuntimeError( 'Vim version is not set.' )
    return VIM_VERSION.major * 100 + VIM_VERSION.minor

  return None


def _MockVimEval( value ): # noqa
  if value == 'g:ycm_neovim_ns_id':
    return 1

  result = _MockVimOptionsEval( value )
  if result is not None:
    return result

  result = _MockVimFunctionsEval( value )
  if result is not None:
    return result

  result = _MockVimBufferEval( value )
  if result is not None:
    return result

  result = _MockVimWindowEval( value )
  if result is not None:
    return result

  result = _MockVimPropEval( value )
  if result is not None:
    return result

  result = _MockVimVersionEval( value )
  if result is not None:
    return result

  match = FNAMEESCAPE_REGEX.search( value )
  if match:
    return match.group( 'filepath' )

  if value == REDIR[ 'variable' ]:
    return REDIR[ 'output' ]

  match = STRDISPLAYWIDTH_REGEX.search( value )
  if match:
    return len( match.group( 'text' ) )

  raise VimError( f'Unexpected evaluation: { value }' )


def _MockWipeoutBuffer( buffer_number ):
  buffers = VIM_MOCK.buffers

  for index, buffer in enumerate( buffers ):
    if buffer.number == buffer_number:
      return buffers.pop( index )


def _MockVimCommand( command ):
  match = BWIPEOUT_REGEX.search( command )
  if match:
    return _MockWipeoutBuffer( int( match.group( 1 ) ) )

  match = REDIR_START_REGEX.search( command )
  if match:
    REDIR[ 'status' ] = True
    REDIR[ 'variable' ] = match.group( 'variable' )
    return

  match = REDIR_END_REGEX.search( command )
  if match:
    REDIR[ 'status' ] = False
    return

  if command == 'unlet ' + REDIR[ 'variable' ]:
    REDIR[ 'variable' ] = ''
    return

  match = LET_REGEX.search( command )
  if match:
    option = match.group( 'option' )
    value = json.loads( match.group( 'value' ) )
    VIM_OPTIONS[ option ] = value
    return

  return DEFAULT


def _MockVimOptions( option ):
  result = VIM_OPTIONS.get( '&' + option )
  if result is not None:
    return result

  return None


class VimBuffer:
  """An object that looks like a vim.buffer object:
   - |name|     : full path of the buffer with symbolic links resolved;
   - |number|   : buffer number;
   - |contents| : list of lines representing the buffer contents;
   - |filetype| : buffer filetype. Empty string if no filetype is set;
   - |modified| : True if the buffer has unsaved changes, False otherwise;
   - |bufhidden|: value of the 'bufhidden' option (see :h bufhidden);
   - |vars|:      dict for buffer-local variables
   - |omnifunc| : omni completion function used by the buffer. Must be a Python
                  function that takes the same arguments and returns the same
                  values as a Vim completion function (:h complete-functions).
                  Example:

                    def Omnifunc( findstart, base ):
                      if findstart:
                        return 5
                      return [ 'a', 'b', 'c' ]"""

  def __init__( self, name,
                      number = 1,
                      contents = [ '' ],
                      filetype = '',
                      modified = False,
                      bufhidden = '',
                      omnifunc = None,
                      visual_start = None,
                      visual_end = None,
                      vars = {} ):
    self.name = os.path.realpath( name ) if name else ''
    self.number = number
    self.contents = contents
    self.filetype = filetype
    self.modified = modified
    self.bufhidden = bufhidden
    self.omnifunc = omnifunc
    self.omnifunc_name = omnifunc.__name__ if omnifunc else ''
    self.changedtick = 1
    self.options = {
     'mod': modified,
     'bh': bufhidden
    }
    self.visual_start = visual_start
    self.visual_end = visual_end
    self.vars = vars # should really be a vim-specific dict-like obj


  def __getitem__( self, index ):
    """Returns the bytes for a given line at index |index|."""
    return self.contents[ index ]


  def __len__( self ):
    return len( self.contents )


  def __setitem__( self, key, value ):
    return self.contents.__setitem__( key, value )


  def GetLines( self ):
    """Returns the contents of the buffer as a list of unicode strings."""
    return [ ToUnicode( x ) for x in self.contents ]


  def mark( self, name ):
    if name == '<':
      return self.visual_start
    if name == '>':
      return self.visual_end
    raise ValueError( f'Unexpected mark: { name }' )


  def __repr__( self ):
    return f"VimBuffer( name = '{ self.name }', number = { self.number } )"


class VimBuffers:
  """An object that looks like a vim.buffers object."""

  def __init__( self, buffers ):
    """|buffers| is a list of VimBuffer objects."""
    self._buffers = buffers


  def __getitem__( self, number ):
    """Emulates vim.buffers[ number ]"""
    for buffer_object in self._buffers:
      if number == buffer_object.number:
        return buffer_object
    raise KeyError( number )


  def __iter__( self ):
    """Emulates for loop on vim.buffers"""
    return iter( self._buffers )


  def pop( self, index ):
    return self._buffers.pop( index )


class VimTabpages:
  def __init__( self, *args ):
    """|buffers| is a list of VimBuffer objects."""
    self._tabpages = []
    self._tabpages.extend( args )


  def __getitem__( self, number ):
    """Emulates vim.buffers[ number ]"""
    for tabpage in self._tabpages:
      if number == tabpage.number:
        return tabpage
    raise KeyError( number )


  def __iter__( self ):
    """Emulates for loop on vim.buffers"""
    return iter( self._tabpages )


class VimWindow:
  """An object that looks like a vim.window object:
    - |number|: number of the window;
    - |buffer_object|: a VimBuffer object representing the buffer inside the
      window;
    - |cursor|: a tuple corresponding to the cursor position."""

  def __init__( self, tabpage, number, buffer_object, cursor = None ):
    self.tabpage = tabpage
    self.number = number
    self.buffer = buffer_object
    self.cursor = cursor
    self.options = {}
    self.vars = {}


  def __repr__( self ):
    return ( f'VimWindow( number = { self.number }, '
                        f'buffer = { self.buffer }, '
                        f'cursor = { self.cursor } )' )


class VimTabpage:
  """An object that looks like a vim.windows object."""

  def __init__( self, number, buffers, cursor ):
    """|buffers| is a list of VimBuffer objects corresponding to the window
    layout. The first element of that list is assumed to be the current window.
    |cursor| is the cursor position of that window."""
    self.number = number
    self.windows = []
    self.windows.append( VimWindow( self, 1, buffers[ 0 ], cursor ) )
    for window_number in range( 1, len( buffers ) ):
      self.windows.append( VimWindow( self,
                                      window_number + 1,
                                      buffers[ window_number ] ) )


  def __getitem__( self, number ):
    """Emulates vim.windows[ number ]"""
    try:
      return self.windows[ number ]
    except IndexError:
      raise IndexError( 'no such window' )


  def __iter__( self ):
    """Emulates for loop on vim.windows"""
    return iter( self.windows )


class VimCurrent:
  """An object that looks like a vim.current object. |current_window| must be a
  VimWindow object."""

  def __init__( self, current_window ):
    self.buffer = current_window.buffer
    self.window = current_window
    self.tabpage = current_window.tabpage
    self.line = self.buffer.contents[ current_window.cursor[ 0 ] - 1 ]


class VimProp:

  def __init__( self,
                prop_type,
                start_line,
                start_column,
                end_line = None,
                end_column = None ):
    current_buffer = VIM_MOCK.current.buffer.number
    self.id = len( VIM_PROPS_FOR_BUFFER[ current_buffer ] ) + 1
    self.prop_type = prop_type
    self.start_line = start_line
    self.start_column = start_column
    self.end_line = end_line if end_line else start_line
    self.end_column = end_column if end_column else start_column


  def __eq__( self, other ):
    return ( self.prop_type == other.prop_type and
             self.start_line == other.start_line and
             self.start_column == other.start_column and
             self.end_line == other.end_line and
             self.end_column == other.end_column )


  def __repr__( self ):
    return ( f"VimProp( prop_type = '{ self.prop_type }',"
             f" start_line = { self.start_line }, "
             f" start_column = { self.start_column },"
             f" end_line = { self.end_line },"
             f" end_column = { self.end_column } )" )


  def __getitem__( self, key ):
    if key == 'type':
      return self.prop_type
    elif key == 'id':
      return self.id
    elif key == 'col':
      return self.start_column
    elif key == 'length':
      return self.end_column - self.start_column


  def get( self, key, default = None ):
    if key == 'type':
      return self.prop_type


class VimSign:

  def __init__( self, line, name, bufnr ):
    self.line = line
    self.name = name
    self.bufnr = bufnr


  def __eq__( self, other ):
    if isinstance( other, dict ):
      other = VimSign( other[ 'lnum' ], other[ 'name' ], other[ 'buffer' ] )
    return ( self.line == other.line and
             self.name == other.name and
             self.bufnr == other.bufnr )


  def __repr__( self ):
    return ( f"VimSign( line = { self.line }, "
                      f"name = '{ self.name }', bufnr = { self.bufnr } )" )


  def __getitem__( self, key ):
    if key == 'group':
      return self.group


@contextlib.contextmanager
def MockVimBuffers( buffers, window_buffers, cursor_position = ( 1, 1 ) ):
  """Simulates the Vim buffers list |buffers| where |current_buffer| is the
  buffer displayed in the current window and |cursor_position| is the current
  cursor position. All buffers are represented by a VimBuffer object."""
  if ( not isinstance( buffers, list ) or
       not all( isinstance( buf, VimBuffer ) for buf in buffers ) ):
    raise RuntimeError( 'First parameter must be a list of VimBuffer objects.' )
  if ( not isinstance( window_buffers, list ) or
       not all( isinstance( buf, VimBuffer ) for buf in window_buffers ) ):
    raise RuntimeError( 'Second parameter must be a list of VimBuffer objects '
                        'representing the window layout.' )
  if len( window_buffers ) < 1:
    raise RuntimeError( 'Second parameter must contain at least one element '
                        'which corresponds to the current window.' )

  with patch( 'vim.buffers', VimBuffers( buffers ) ):
    with patch( 'vim.tabpages', VimTabpages(
      VimTabpage( 1, window_buffers, cursor_position ) ) ) as tabpages:
      with patch( 'vim.windows', tabpages[ 1 ] ) as windows:
        with patch( 'vim.current', VimCurrent( windows[ 0 ] ) ):
          yield VIM_MOCK


def MockVimModule():
  """The 'vim' module is something that is only present when running inside the
  Vim Python interpreter, so we replace it with a MagicMock for tests. If you
  need to add additional mocks to vim module functions, then use 'patch' from
  mock module, to ensure that the state of the vim mock is returned before the
  next test. That is:

    from ycm.tests.test_utils import MockVimModule
    from unittest.mock import patch

    # Do this once
    MockVimModule()

    @patch( 'vim.eval', return_value='test' )
    @patch( 'vim.command', side_effect=ValueError )
    def test( vim_command, vim_eval ):
      # use vim.command via vim_command, e.g.:
      vim_command.assert_has_calls( ... )

  Failure to use this approach may lead to unexpected failures in other
  tests."""

  VIM_MOCK.command = MagicMock( side_effect = _MockVimCommand )
  VIM_MOCK.eval = MagicMock( side_effect = _MockVimEval )
  VIM_MOCK.error = VimError
  VIM_MOCK.options = MagicMock()
  VIM_MOCK.options.__getitem__.side_effect = _MockVimOptions
  sys.modules[ 'vim' ] = VIM_MOCK

  return VIM_MOCK


class VimError( Exception ):

  def __init__( self, code ):
    self.code = code


  def __str__( self ):
    return repr( self.code )


class ExtendedMock( MagicMock ):
  """An extension to the MagicMock class which adds the ability to check that a
  callable is called with a precise set of calls in a precise order.

  Example Usage:
    from ycm.tests.test_utils import ExtendedMock
    @patch( 'test.testing', new_callable = ExtendedMock, ... )
    def my_test( test_testing ):
      ...
  """

  def assert_has_exact_calls( self, calls, any_order = False ):
    contains = contains_inanyorder if any_order else contains_exactly
    assert_that( self.call_args_list, contains( *calls ) )
    assert_that( self.call_count, equal_to( len( calls ) ) )


def ExpectedFailure( reason, *exception_matchers ):
  """Defines a decorator to be attached to tests. This decorator
  marks the test as being known to fail, e.g. where documenting or exercising
  known incorrect behaviour.

  The parameters are:
    - |reason| a textual description of the reason for the known issue. This
               is used for the skip reason
    - |exception_matchers| additional arguments are hamcrest matchers to apply
                 to the exception thrown. If the matchers don't match, then the
                 test is marked as error, with the original exception.

  If the test fails (for the correct reason), then it is marked as skipped.
  If it fails for any other reason, it is marked as failed.
  If the test passes, then it is also marked as failed."""
  def decorator( test ):
    @functools.wraps( test )
    def Wrapper( *args, **kwargs ):
      try:
        test( *args, **kwargs )
      except Exception as test_exception:
        # Ensure that we failed for the right reason
        test_exception_message = ToUnicode( test_exception )
        try:
          for matcher in exception_matchers:
            assert_that( test_exception_message, matcher )
        except AssertionError:
          # Failed for the wrong reason!
          import traceback
          print( 'Test failed for the wrong reason: ' + traceback.format_exc() )
          # Real failure reason is the *original* exception, we're only trapping
          # and ignoring the exception that is expected.
          raise test_exception

        # Failed for the right reason
        skip( reason )
      else:
        raise AssertionError( f'Test was expected to fail: { reason }' )
    return Wrapper

  return decorator



# Copyright (C) 2016-2017 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm.client.base_request import BaseRequest, BuildRequestData


class DebugInfoRequest( BaseRequest ):
  def __init__( self, extra_data = None ):
    super( DebugInfoRequest, self ).__init__()
    self._extra_data = extra_data
    self._response = None


  def Start( self ):
    request_data = BuildRequestData()
    if self._extra_data:
      request_data.update( self._extra_data )
    self._response = self.PostDataToHandler( request_data,
                                             'debug_info',
                                             display_message = False )


  def Response( self ):
    return self._response


def FormatDebugInfoResponse( response ):
  if not response:
    return 'Server errored, no debug info from server\n'
  message = _FormatYcmdDebugInfo( response )
  completer = response[ 'completer' ]
  if completer:
    message += _FormatCompleterDebugInfo( completer )
  return message


def _FormatYcmdDebugInfo( ycmd ):
  python = ycmd[ 'python' ]
  clang = ycmd[ 'clang' ]
  message = (
    f'Server Python interpreter: { python[ "executable" ] }\n'
    f'Server Python version: { python[ "version" ] }\n'
    f'Server has Clang support compiled in: { clang[ "has_support" ] }\n'
    f'Clang version: { clang[ "version" ] }\n' )
  extra_conf = ycmd[ 'extra_conf' ]
  extra_conf_path = extra_conf[ 'path' ]
  if not extra_conf_path:
    message += 'No extra configuration file found\n'
  elif not extra_conf[ 'is_loaded' ]:
    message += ( 'Extra configuration file found but not loaded\n'
                 f'Extra configuration path: { extra_conf_path }\n' )
  else:
    message += ( 'Extra configuration file found and loaded\n'
                 f'Extra configuration path: { extra_conf_path }\n' )
  return message


def _FormatCompleterDebugInfo( completer ):
  message = f'{ completer[ "name" ] } completer debug information:\n'
  for server in completer[ 'servers' ]:
    name = server[ 'name' ]
    if server[ 'is_running' ]:
      address = server[ 'address' ]
      port = server[ 'port' ]
      if address and port:
        message += f'  { name } running at: http://{ address }:{ port }\n'
      else:
        message += f'  { name } running\n'
      message += f'  { name } process ID: { server[ "pid" ] }\n'
    else:
      message += f'  { name } not running\n'
    message += f'  { name } executable: { server[ "executable" ] }\n'
    logfiles = server[ 'logfiles' ]
    if logfiles:
      message += f'  { name } logfiles:\n'
      for logfile in logfiles:
        message += f'    { logfile }\n'
    else:
      message += '  No logfiles available\n'
    if 'extras' in server:
      for extra in server[ 'extras' ]:
        message += f'  { name } { extra[ "key" ] }: { extra[ "value" ] }\n'
  for item in completer[ 'items' ]:
    message += f'  { item[ "key" ].capitalize() }: { item[ "value" ] }\n'
  return message


def SendDebugInfoRequest( extra_data = None ):
  request = DebugInfoRequest( extra_data )
  # This is a blocking call.
  request.Start()
  return request.Response()



# Comment
def Test_OneLine():
  """This is the one line output."""
  pass

def Test_MultiLine():
  """This is the one line output.
  This is second line."""
  pass

def Main():
  Test_OneLine()
  Test_MultiLine()


def Really_Long_Method( which, has, some, param, that, take, the, whole, line ):
  """Lorem ipsum dolor sit amet, consectetur adipiscing elit. Vestibulum egestas libero urna, vel sagittis felis condimentum in. Nulla arcu eros, aliquet vel mollis vitae, semper eu ex. Donec posuere quam et ornare sagittis. Curabitur nunc ex, fringilla quis lorem sed, dignissim congue felis. Integer vestibulum ac elit vel blandit. Nam non dui urna. Integer eu semper massa. Nullam ac elit interdum, aliquet elit nec, porttitor orci. Duis tempus justo lorem, ac fringilla ante viverra egestas. Etiam eleifend enim ac libero dapibus, quis condimentum lectus tristique. Fusce feugiat, lorem et faucibus eleifend, ipsum nisi maximus justo, at consectetur ligula leo vitae justo."""
  # Really long one-line
  pass


def Really_Long_Method_2():
  """Lorem ipsum dolor sit amet, consectetur adipiscing elit. Vestibulum egestas
  libero urna, vel sagittis felis condimentum in. Nulla arcu eros, aliquet vel
  mollis vitae, semper eu ex. Donec posuere quam et ornare sagittis. Curabitur
  nunc ex, fringilla quis lorem sed, dignissim congue felis. Integer vestibulum
  ac elit vel blandit. Nam non dui urna. Integer eu semper massa. Nullam ac elit
  interdum, aliquet elit nec, porttitor orci. Duis tempus justo lorem, ac
  fringilla ante viverra egestas. Etiam eleifend enim ac libero dapibus, quis
  condimentum lectus tristique. Fusce feugiat, lorem et faucibus eleifend, ipsum
  nisi maximus justo, at consectetur ligula leo vitae justo."""
  # Really long one para
  pass


def Moan():
  Really_Long_Method()
  Really_Long_Method_2()



# Copyright (C) 2017 YouCompleteMe Contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy
from hamcrest import assert_that, contains_string, equal_to
from unittest import TestCase

from ycm.client.debug_info_request import FormatDebugInfoResponse


GENERIC_RESPONSE = {
  'clang': {
    'has_support': True,
    'version': 'Clang version'
  },
  'completer': {
    'items': [
      {
        'key': 'key',
        'value': 'value'
      }
    ],
    'name': 'Completer name',
    'servers': [
      {
        'address': '127.0.0.1',
        'executable': '/path/to/executable',
        'extras': [
          {
            'key': 'key',
            'value': 'value'
          }
        ],
        'is_running': True,
        'logfiles': [
          '/path/to/stdout/logfile',
          '/path/to/stderr/logfile'
        ],
        'name': 'Server name',
        'pid': 12345,
        'port': 1234
      }
    ]
  },
  'extra_conf': {
    'is_loaded': False,
    'path': '/path/to/extra/conf'
  },
  'python': {
    'executable': '/path/to/python/interpreter',
    'version': 'Python version'
  }
}


class DebugInfoRequestTest( TestCase ):
  def test_FormatDebugInfoResponse_NoResponse( self ):
    assert_that(
      FormatDebugInfoResponse( None ),
      equal_to( 'Server errored, no debug info from server\n' )
    )


  def test_FormatDebugInfoResponse_NoExtraConf( self ):
    response = deepcopy( GENERIC_RESPONSE )
    response[ 'extra_conf' ].update( {
      'is_loaded': False,
      'path': None
    } )
    assert_that(
      FormatDebugInfoResponse( response ),
      contains_string(
        'No extra configuration file found\n'
      )
    )


  def test_FormatDebugInfoResponse_ExtraConfFoundButNotLoaded( self ):
    response = deepcopy( GENERIC_RESPONSE )
    response[ 'extra_conf' ].update( {
      'is_loaded': False,
      'path': '/path/to/extra/conf'
    } )
    assert_that(
      FormatDebugInfoResponse( response ),
      contains_string(
        'Extra configuration file found but not loaded\n'
        'Extra configuration path: /path/to/extra/conf\n'
      )
    )


  def test_FormatDebugInfoResponse_ExtraConfFoundAndLoaded( self ):
    response = deepcopy( GENERIC_RESPONSE )
    response[ 'extra_conf' ].update( {
      'is_loaded': True,
      'path': '/path/to/extra/conf'
    } )
    assert_that(
      FormatDebugInfoResponse( response ),
      contains_string(
        'Extra configuration file found and loaded\n'
        'Extra configuration path: /path/to/extra/conf\n'
      )
    )


  def test_FormatDebugInfoResponse_Completer_ServerRunningWithHost( self ):
    response = deepcopy( GENERIC_RESPONSE )
    assert_that(
      FormatDebugInfoResponse( response ),
      contains_string(
        'Completer name completer debug information:\n'
        '  Server name running at: http://127.0.0.1:1234\n'
        '  Server name process ID: 12345\n'
        '  Server name executable: /path/to/executable\n'
        '  Server name logfiles:\n'
        '    /path/to/stdout/logfile\n'
        '    /path/to/stderr/logfile\n'
        '  Server name key: value\n'
        '  Key: value\n'
      )
    )


  def test_FormatDebugInfoResponse_Completer_ServerRunningWithoutHost( self ):
    response = deepcopy( GENERIC_RESPONSE )
    response[ 'completer' ][ 'servers' ][ 0 ].update( {
      'address': None,
      'port': None
    } )
    assert_that(
      FormatDebugInfoResponse( response ),
      contains_string(
        'Completer name completer debug information:\n'
        '  Server name running\n'
        '  Server name process ID: 12345\n'
        '  Server name executable: /path/to/executable\n'
        '  Server name logfiles:\n'
        '    /path/to/stdout/logfile\n'
        '    /path/to/stderr/logfile\n'
        '  Server name key: value\n'
        '  Key: value\n'
      )
    )


  def test_FormatDebugInfoResponse_Completer_ServerNotRunningWithNoLogfiles(
      self ):
    response = deepcopy( GENERIC_RESPONSE )
    response[ 'completer' ][ 'servers' ][ 0 ].update( {
      'is_running': False,
      'logfiles': []
    } )
    assert_that(
      FormatDebugInfoResponse( response ),
      contains_string(
        'Completer name completer debug information:\n'
        '  Server name not running\n'
        '  Server name executable: /path/to/executable\n'
        '  No logfiles available\n'
        '  Server name key: value\n'
        '  Key: value\n'
      )
    )
